import json
from logging import Logger
from xml.etree import cElementTree as eTree

from application.model import BenefitAppliers
from application.model.repository.benefit_applier.exception import BenefitApplierNotFound
from application.model.repository.benefit_applier import DefaultBenefitApplierRepository
from typing import Optional, Dict, List

BENEFIT_APPLIERS_REPOSITORIES = None  # type: Optional[Dict[BenefitAppliers: DefaultBenefitApplierRepository]]


def find_benefit_applier(
    pos_id,     # type: int
    coupon_id,  # type: str
):
    # type: (...) -> BenefitAppliers

    for benefit_applier in sorted(BENEFIT_APPLIERS_REPOSITORIES, reverse=True):
        if BENEFIT_APPLIERS_REPOSITORIES[benefit_applier].check_coupon_id(pos_id, coupon_id):
            return benefit_applier

    raise BenefitApplierNotFound("NDISCOUNT_BENEFIT_NOT_FOUND")


def find_benefit_id_by_custom_property(order_picture):
    # type: (eTree) -> Optional[str]

    for benefit_applier in BENEFIT_APPLIERS_REPOSITORIES:
        benefit_id = BENEFIT_APPLIERS_REPOSITORIES[benefit_applier].get_benefit_id_by_custom_property(order_picture)
        if benefit_id:
            return benefit_applier, benefit_id


def find_order_benefits_id(logger, order_picture, validate_voided_lines=False):
    # type: (Logger, eTree, Optional[bool]) -> Optional[List[str]]

    order_id = order_picture.get("orderId")
    logger.info("Finding order benefits to orderId: {}".format(order_id))

    order_benefits_id = []
    try:
        benefit_prop = order_picture.find(".//OrderProperty[@key='BENEFIT_LIST']")
        if benefit_prop is None:
            return order_benefits_id

        voided_sale_lines = _get_removed_sale_lines(order_picture)

        applied_benefits = json.loads(benefit_prop.get("value", ""))
        benefit_list = json.loads(applied_benefits.get("benefitList"))

        for applied_benefit in benefit_list:
            if validate_voided_lines:
                added_sale_lines = applied_benefit.get("added_sale_lines")
                has_voided_sale_lines = [sale_line for sale_line in added_sale_lines if sale_line in voided_sale_lines]

                if not has_voided_sale_lines:
                    continue

            benefit = applied_benefit.get("benefit")
            if benefit:
                benefit_id = benefit.get("id")
            else:
                continue

            if benefit_id not in order_benefits_id:
                order_benefits_id.append(benefit_id)

        return order_benefits_id
    finally:
        logger.info("Found [{}] benefits on orderId: {}".format(len(order_benefits_id), order_id))


def _get_removed_sale_lines(order_picture):
    # type: (eTree) -> List

    sale_lines = order_picture.findall(".//SaleLine")
    removed_sale_lines = []
    for sale_line in sale_lines:
        if sale_line.get("level") == "0" and sale_line.get("qty") == "0":
            removed_sale_lines.append(int(sale_line.get("lineNumber")))

    return removed_sale_lines


def fill_benefit_appliers_repositories(benefit_appliers_repositories):
    # type: (Dict[BenefitAppliers: DefaultBenefitApplierRepository]) -> None

    global BENEFIT_APPLIERS_REPOSITORIES

    BENEFIT_APPLIERS_REPOSITORIES = benefit_appliers_repositories


def get_originator_pos_id(order_picture):
    # type: (eTree) -> int

    originator_pos_id = order_picture.get("originatorId")
    if originator_pos_id is not None:
        return int(originator_pos_id.split("POS")[-1])



from abc import ABCMeta, abstractmethod

from commons.util import Translator


class Formatter(object):
    _metaclass__ = ABCMeta

    @abstractmethod
    def format_report(self, pick_list_dto, translator):
        # type: (object, Translator) -> str

        raise NotImplementedError()



from ._ApplyCouponContext import ApplyCouponContext


class ApplyCouponRequest(object):
    def __init__(
        self,
        coupon_id,  # type: str
        context,    # type: ApplyCouponContext
    ):
        # type: (...) -> None

        self.coupon_id = coupon_id
        self.context = context



from logging import Logger
from typing import Dict

from interactor.route.store.get_config.dto import GetStoreConfigRequest
from model.config import Config
from model.repository import PosRepositoryModel
from validation_util import validate_api_key


class GetStoreConfigInteractor(object):

    def __init__(
            self,
            logger,         # type: Logger
            config,         # type: Config
            pos_repository  # type: PosRepositoryModel
    ):
        # type: (...) -> None

        self.logger = logger
        self.config = config
        self.pos_repository = pos_repository

    def execute(self, request):
        # type: (GetStoreConfigRequest) -> Dict

        pos_id = validate_api_key(
            request=request,
            secure_api_key_list=self.config.api_key_mapping
        )

        return self.pos_repository.get_static_config(pos_id)



from application.interactor import BenefitRetryInteractor
from application.model import BaseThread
from application.model.configuration import Configurations


class BenefitRetryManager(BaseThread):
    def __init__(self, configs, interactor):
        # type: (Configurations, BenefitRetryInteractor) -> None

        super(BenefitRetryManager, self).__init__()
        self.configs = configs
        self.logger = self.configs.logger
        self.interactor = interactor

    def run(self):
        # type: () -> None
        
        while self.running:
            try:
                pending_benefits = self.interactor.get_pending_operations_benefits()
                if not pending_benefits:
                    continue

                benefits_ids = [benefit.benefit_id for benefit in pending_benefits]
                self.logger.info("Found [{}] benefits to retry. Benefits: {}".format(len(benefits_ids), benefits_ids))
                self.interactor.perform_operations_on_pending_benefits(pending_benefits)
            except Exception as _:
                self.logger.exception("Not expected exception")
            finally:
                self.sleep(self.configs.retry_frequency)



from application.domain import OrderManager, SaleLine  # noqa
from sysactions import get_posot, get_model


class MsgBusOrderManager(OrderManager):

    def recall_order(self, pos_id, order_id):
        model = get_model(pos_id)
        pos_ot = get_posot(model)
        pos_ot.recallOrder(pos_id, order_id)

    def store_order(self, pos_id):
        model = get_model(pos_id)
        pos_ot = get_posot(model)
        pos_ot.storeOrder(pos_id)

    def change_line_seat(self, pos_id, sale_line, seat_number):
        # type: (str, SaleLine, str) -> None

        model = get_model(pos_id)
        pos_ot = get_posot(model)

        pos_ot.setOrderCustomProperty("seat",
                                      str(seat_number),
                                      sale_line.order_id,
                                      sale_line.line_number,
                                      sale_line.item_id,
                                      sale_line.level,
                                      sale_line.part_code)

    def get_current_order_state(self, pos_id):
        model = get_model(pos_id)

        order_node = model.find("CurrentOrder/Order")
        if order_node is not None:
            return order_node.get("state")
        return None



# -*- coding: utf-8 -*-

from sqlite3 import Connection

from msgbus import MBEasyContext
from old_helper import BaseRepository
from typing import List, Dict


class OrderRepository(BaseRepository):

    def __init__(
        self,
        mb_context,  # type: MBEasyContext
        pos_list,   # type: List[int]
    ):
        # type: (...) -> None
        super(OrderRepository, self).__init__(mb_context)
        self.pos_list = pos_list

    def get_orders_amount(
        self,
        order_ids,     # type: List[str]
    ):
        # type: (...) -> Dict
        orders = {}

        def inner_func(conn):
            # type: (Connection) -> Dict

            query = self._OrdersValueQuery.format(",".join(order_ids))
            for x in conn.select(query):
                order_id = x.get_entry("OrderId")
                amount = x.get_entry("TotalGross")
                tip = x.get_entry("Tip")
                if order_id not in orders:
                    orders[order_id] = {"amount": 0, "tip": 0}

                orders[order_id]["amount"] += float(amount)
                orders[order_id]["tip"] += float(tip)

            return orders

        self.execute_parallel_in_all_databases_returning_flat_list(inner_func, self.pos_list)
        return orders

    _OrdersValueQuery = """
        SELECT 
            OrderId,
            TotalGross,
            Tip
        FROM Orders
        WHERE OrderId in ({}) AND StateId IN (2, 3) 
    """



from unicodedata import normalize

from report import Formatter
from helper import (
    remove_accents,
    custom_encode_cp860,
)
from typing import Optional

from _CommandFactory import CommandFactory
from report import Report

ESC = b'\x1b'


class EscPosFormatter(Formatter):
    def __init__(self, encoding="cp860", command_factory=None):
        # type: (str, CommandFactory) -> None
        self.encoding = encoding
        self.command_factory = command_factory

    def format(
        self,
        report,                 # type: Report
        allow_accents=False,    # type: Optional[bool]
        is_mobile=False,        # type: Optional[bool]
    ):
        # type: (...) -> str

        body = self.create_report_body(report.get_parts())
        if body == "":
            return body

        content = self._get_page_command() + body
        if is_mobile:
            return content.encode('utf-8')
        if not allow_accents:
            return remove_accents(text=content).encode(self.encoding)
        return custom_encode_cp860(text=content)

    def create_report_body(self, parts):
        body = ""
        for part in parts:
            if part.commands is not None:
                for command in part.commands:
                    body += self.command_factory.get_processor(command).get_bytes(command)
            if part.text is not None:
                body += part.text
        return body

    def _get_page_command(self):
        if self.encoding == "cp850":
            return ESC + 't' + '\x02'
        else:
            return ESC + 't' + '\x03'



# encoding: utf-8
from mw import MwOperatorNameRetriever

from openedtablesreport import OpenedTablesReport
from openedtablesreport.repository import (
    OrderRepository,
    TableRepository,
)
from utils import get_pos_list


class ReportGenerator(object):
    @staticmethod
    def generate(mb_context, pos_id):
        operator_retriever = MwOperatorNameRetriever(
            mb_context
        )
        table_repository = TableRepository(
            mb_context,
            pos_id
        )
        order_repository = OrderRepository(
            mb_context,
            get_pos_list(mb_context)
        )
        opened_tables_report_generator = OpenedTablesReport(
            table_repository,
            order_repository,
            pos_id,
            operator_retriever
        )
        return opened_tables_report_generator.generate_opened_tables_report(pos_id)




from abc import ABCMeta, abstractmethod

from model.electronic_message import ElectronicMessage


class ApiServiceModel(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def send_electronic_messages(self, electronic_messages):
        # type: (ElectronicMessage) -> None

        raise NotImplementedError



from model.loyalty import Benefit


class BenefitInfo(object):
    def __init__(
        self,
        notify,     # type: bool
        benefit,    # type: Benefit
    ):
        # type: (...) -> None

        self.notify = notify
        self.benefit = benefit



from ._GetLoyaltyCustomerInfoProcessor import GetLoyaltyCustomerInfoProcessor
from ._BurnVoucherProcessor import BurnVoucherProcessor
from ._UnlockVoucherProcessor import UnlockVoucherProcessor
from ._LoyaltyAvailableProcessor import LoyaltyAvailableProcessor
from ._LoyaltyIdentifierProcessor import LoyaltyIdentifierProcessor



from logging import Logger
from services.model import CleanDatabasesRequest
from externals.repository import AlertsRepository
from services.model import Configs


class CleanDatabasesInteractor(object):
    def __init__(
        self,
        alerts_repository,  # type: AlertsRepository
        configs,            # type: Configs
        logger,             # type: Logger
    ):
        # type: (...) -> None

        self.alerts_repository = alerts_repository
        self.configs = configs
        self.logger = logger

    def process_request(
        self,
        request,  # type: CleanDatabasesRequest
    ):
        # type: (...) -> None

        try:
            outdated_event_days = request.default_outdated_events_days
            if "OutdatedEventDays" in self.configs.alerts:
                outdated_event_days = int(self.configs.alerts["OutdatedEventDays"])

            self.alerts_repository.delete_outdated_events(range_days=outdated_event_days)
            self.alerts_repository.vacuum_database()
        except Exception as e:
            self.logger.exception(e.message)



from datetime import datetime
from xml.etree import cElementTree as Etree

from interactor.event import EventInteractor
from model.log import LogEntry
from processor.event import EventProcessor


def _get_log_entry_from_pos_event(
        data,           # type: str
        event_type,     # type: str
):
    event = Etree.fromstring(data)
    event_data = event[0] if "Event" in event.tag else event
    created_at = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")

    return LogEntry(
            pos_id=event_data.get("posId", event_data.get("posid")) or -1,
            business_period=event_data.get("businessPeriod", datetime.utcnow().strftime("%Y%m%d")),
            event_type=event_type,
            created_at_gmt=created_at,
            event_data=data
    )


class GenericEventProcessor(
        EventProcessor  # type: EventInteractor
):

    def parse_data(self, data, event_type):
        # type: (str, str) -> LogEntry

        return _get_log_entry_from_pos_event(data, event_type)



import logging

from application.repository.api import APIRepository
from typing import Optional

logger = logging.getLogger(__name__)


class GetLoyaltyCustomerInfoInteractor(object):
    def __init__(self, api_repository):
        # type: (APIRepository) -> None

        self.api_repository = api_repository

    def get_loyalty_customer_info(
        self,                 # type: GetLoyaltyCustomerInfoInteractor
        loyalty_customer_id,  # type: str
    ):
        # type: (...) -> Optional[str]

        logger.info("Requesting customer loyaltySystemId")
        loyalty_id = self.api_repository.retrieve_loyalty_customer_info(loyalty_customer_id=loyalty_customer_id)
        logger.info("Retrieved customer loyalty data: {}".format(loyalty_id))
        return loyalty_id



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

# Copyright (C) 2003-2007, 2009-2011 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose with or without fee is hereby granted,
# provided that the above copyright notice and this permission notice
# appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import dns.rdtypes.mxbase


class AFSDB(dns.rdtypes.mxbase.UncompressedDowncasingMX):

    """AFSDB record

    @ivar subtype: the subtype value
    @type subtype: int
    @ivar hostname: the hostname name
    @type hostname: dns.name.Name object"""

    # Use the property mechanism to make "subtype" an alias for the
    # "preference" attribute, and "hostname" an alias for the "exchange"
    # attribute.
    #
    # This lets us inherit the UncompressedMX implementation but lets
    # the caller use appropriate attribute names for the rdata type.
    #
    # We probably lose some performance vs. a cut-and-paste
    # implementation, but this way we don't copy code, and that's
    # good.

    def get_subtype(self):
        return self.preference

    def set_subtype(self, subtype):
        self.preference = subtype

    subtype = property(get_subtype, set_subtype)

    def get_hostname(self):
        return self.exchange

    def set_hostname(self, hostname):
        self.exchange = hostname

    hostname = property(get_hostname, set_hostname)



from enum import Enum


class OrderStateEnum(Enum):
    UNDEFINED = 0
    IN_PROGRESS = 1
    STORED = 2
    TOTALED = 3
    VOIDED = 4
    PAID = 5
    RECALLED = 6
    SYSTEM_VOIDED = 7
    ABANDONED = 8



from old_helper import BaseRepository
from persistence import Connection
from typing import List, Tuple

from ..model import Order


class FiscalRepository(BaseRepository):
    def __init__(self, mbcontext):
        super(FiscalRepository, self).__init__(mbcontext)

    def get_card_brands(self, paid_orders):
        # type: (List[Order]) -> List[Tuple[unicode, int, float, int]]
        def inner_func(conn):
            # type: (Connection) -> List[Tuple[unicode, int, float, int]]
            order_list = map(lambda order: order.order_id, paid_orders)
            order_list = ','.join(map(str, order_list))
            query = self._CardBrandsQuery
            query = query.format(order_list)
            return [((x.get_entry(0)).decode('utf-8'), int(x.get_entry(1)), float(x.get_entry(2)), int(x.get_entry(3))) for x in conn.select(query)]

        return self.execute_with_connection(inner_func, None, "FiscalPersistence")

    _CardBrandsQuery = """SELECT
                          CASE WHEN bc.Descricao IS NOT NULL THEN bc.Descricao ELSE 'OUTROS' END AS Brand, pd.Type as Type, SUM(pd.Amount) as Amount, COUNT(pd.OrderId)
                          FROM fiscal.PaymentData pd
                          LEFT JOIN fiscal.BandeiraCartao bc on pd.Bandeira = bc.Bandeira
                          WHERE pd.Type IN (1,2,3) AND pd.OrderId IN ({0})
                          GROUP BY bc.Descricao, pd.Type
                          ORDER BY Type ASC, Amount DESC"""



# -*- coding: utf-8 -*-
# vi:tabstop=4:expandtab:sw=4
"""Transliterate Unicode text into plain 7-bit ASCII.

Example usage:
>>> from unidecode import unidecode
>>> unidecode(u"\u5317\u4EB0")
"Bei Jing "

The transliteration uses a straightforward map, and doesn't have alternatives
for the same character based on language, position, or anything else.

In Python 3, a standard string object will be returned. If you need bytes, use:
>>> unidecode("Κνωσός").encode("ascii")
b'Knosos'
"""
import warnings
from sys import version_info

Cache = {}


def _warn_if_not_unicode(string):
    if version_info[0] < 3 and not isinstance(string, unicode):
        warnings.warn(  "Argument %r is not an unicode object. "
                        "Passing an encoded string will likely have "
                        "unexpected results." % (type(string),),
                        RuntimeWarning, 2)


def unidecode_expect_ascii(string):
    """Transliterate an Unicode object into an ASCII string

    >>> unidecode(u"\u5317\u4EB0")
    "Bei Jing "

    This function first tries to convert the string using ASCII codec.
    If it fails (because of non-ASCII characters), it falls back to
    transliteration using the character tables.

    This is approx. five times faster if the string only contains ASCII
    characters, but slightly slower than using unidecode directly if non-ASCII
    chars are present.
    """

    _warn_if_not_unicode(string)
    try:
        bytestring = string.encode('ASCII')
    except UnicodeEncodeError:
        return _unidecode(string)
    if version_info[0] >= 3:
        return string
    else:
        return bytestring

def unidecode_expect_nonascii(string):
    """Transliterate an Unicode object into an ASCII string

    >>> unidecode(u"\u5317\u4EB0")
    "Bei Jing "
    """

    _warn_if_not_unicode(string)
    return _unidecode(string)

unidecode = unidecode_expect_ascii

def _unidecode(string):
    retval = []

    for char in string:
        codepoint = ord(char)

        if codepoint < 0x80: # Basic ASCII
            retval.append(str(char))
            continue
        
        if codepoint > 0xeffff:
            continue # Characters in Private Use Area and above are ignored

        if 0xd800 <= codepoint <= 0xdfff:
            warnings.warn(  "Surrogate character %r will be ignored. "
                            "You might be using a narrow Python build." % (char,),
                            RuntimeWarning, 2)

        section = codepoint >> 8   # Chop off the last two hex digits
        position = codepoint % 256 # Last two hex digits

        try:
            table = Cache[section]
        except KeyError:
            try:
                mod = __import__('unidecode.x%03x'%(section), globals(), locals(), ['data'])
            except ImportError:
                Cache[section] = None
                continue   # No match: ignore this character and carry on.

            Cache[section] = table = mod.data

        if table and len(table) > position:
            retval.append( table[position] )

    return ''.join(retval)



from ._SetOrderStateProcessor import SetOrderStateProcessor
from ._ToggleTagLineProcessor import ToggleTagLineProcessor
from ._UndoProcessor import UndoProcessor
from ._RefreshViewProcessor import RefreshViewProcessor
from ._EnablePathProcessor import EnablePathProcessor
from ._DisablePathProcessor import DisablePathProcessor
from ._GetOrderXmlProcessor import GetOrderXmlProcessor




""" ESC/POS configuration manager.

This module contains the implentations of abstract base class :py:class:`Config`.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import appdirs
import yaml

from . import printer
from . import exceptions


class Config(object):
    """  Configuration handler class.

    This class loads configuration from a default or specificed directory. It
    can create your defined printer and return it to you.
    """
    _app_name = 'python-escpos'
    _config_file = 'config.yaml'

    def __init__(self):
        """ Initialize configuration.

        Remember to add anything that needs to be reset between configurations
        to self._reset_config
        """
        self._has_loaded = False
        self._printer = None

        self._printer_name = None
        self._printer_config = None

    def _reset_config(self):
        """ Clear the loaded configuration.

        If we are loading a changed config, we don't want to have leftover
        data.
        """
        self._has_loaded = False
        self._printer = None

        self._printer_name = None
        self._printer_config = None

    def load(self, config_path=None):
        """ Load and parse the configuration file using pyyaml

        :param config_path: An optional file path, file handle, or byte string
            for the configuration file.

        """

        self._reset_config()

        if not config_path:
            config_path = os.path.join(
                appdirs.user_config_dir(self._app_name),
                self._config_file
            )

        try:
            # First check if it's file like. If it is, pyyaml can load it.
            # I'm checking type instead of catching exceptions to keep the
            # exception handling simple
            if hasattr(config_path, 'read'):
                config = yaml.safe_load(config_path)
            else:
                # If it isn't, it's a path. We have to open it first, otherwise
                # pyyaml will try to read it as yaml
                with open(config_path, 'rb') as config_file:
                    config = yaml.safe_load(config_file)
        except EnvironmentError:
            raise exceptions.ConfigNotFoundError('Couldn\'t read config at {config_path}'.format(
                config_path=str(config_path),
            ))
        except yaml.YAMLError:
            raise exceptions.ConfigSyntaxError('Error parsing YAML')

        if 'printer' in config:
            self._printer_config = config['printer']
            self._printer_name = self._printer_config.pop('type').title()

            if not self._printer_name or not hasattr(printer, self._printer_name):
                raise exceptions.ConfigSyntaxError(
                    'Printer type "{printer_name}" is invalid'.format(
                        printer_name=self._printer_name,
                    )
                )

        self._has_loaded = True

    def printer(self):
        """ Returns a printer that was defined in the config, or throws an
        exception.

        This method loads the default config if one hasn't beeen already loaded.

        """
        if not self._has_loaded:
            self.load()

        if not self._printer_name:
            raise exceptions.ConfigSectionMissingError('printer')

        if not self._printer:
            # We could catch init errors and make them a ConfigSyntaxError,
            # but I'll just let them pass
            self._printer = getattr(printer, self._printer_name)(**self._printer_config)

        return self._printer



from abc import (
    ABCMeta,
    abstractmethod
)
from typing import Any


class NavigationServiceModel(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def get_navigation_descriptions(
        self,
    ):
        # type: (...) -> Any

        raise NotImplementedError



# encoding: utf-8
from ._JustificationHeaderDto import JustificationHeaderDto
from ._JustificationBodyDto import JustificationBodyDto


class JustificationGeneratorDto(object):
    def __init__(
        self,
        header,     # type: JustificationHeaderDto
        body,       # type: JustificationBodyDto
    ):
        # type: (...) -> None

        self.header = header
        self.body = body



class ListenedEvents(object):

    # ElectronicPayment
    ELECTRONIC_PAYMENT_SHOW_CHOICES = "SHOW_CHOICES"
    ELECTRONIC_PAYMENT_SHOW_MESSAGE = "SHOW_MESSAGE"
    ELECTRONIC_PAYMENT_SHOW_QUESTION = "SHOW_QUESTION"
    ELECTRONIC_PAYMENT_SHOW_QRCODE = "SHOW_QRCODE"
    ELECTRONIC_PAYMENT_SHOW_INPUT = "SHOW_INPUT"
    ELECTRONIC_PAYMENT_CLOSE_DIALOGS = "CLOSE_DIALOGS"



from httpmessageprocessorutil.automapped import auto_map
from typing import Optional


@auto_map({"system_api_key": {"name": "X-Api-Key", "location": "header", "type": str},
           "pos_id": {"name": "posId", "type": str},
           "price_list": {"name": "priceList", "type": Optional[str]},
           "sale_type": {"name": "saleType", "type": Optional[str]}})
class OrderCreateRequest(object):
    def __init__(self, system_api_key, pos_id, price_list="", sale_type=""):
        # type: (str, str, Optional[str], Optional[str]) -> None

        self.system_api_key = system_api_key
        self.pos_id = pos_id
        self.price_list = price_list
        self.sale_type = sale_type



import json
from abc import abstractmethod
from xml.etree import cElementTree as eTree

from msgbus import TK_SYS_NAK

from application.model.configuration import Configurations
from application.model.customexception import ErrorApplyingBenefit
from mbcontextmessagehandler import MbContextMessageBus
from messagebus import Message, DataType
from typing import Optional
from ... import NDiscountTokens


class DefaultBenefitApplierRepository(object):

    def __init__(self, configs, message_bus):
        # type: (Configurations, MbContextMessageBus) -> None

        self.logger = configs.logger
        self.configs = configs
        self.message_bus = message_bus

    def add_benefit(
            self,
            current_pos_id,             # type: int
            benefit_id,                 # type: str
            benefit,                    # type: str
            benefit_applier,            # type: str
            tender_id,                  # type: str
            user_auth_level             # type: bool
    ):
        # type: (...) -> None

        data = json.dumps(dict(
                posId=current_pos_id,
                benefitId=benefit_id,
                benefit=benefit,
                benefit_applier=benefit_applier,
                tender_id=tender_id,
                user_auth_level=user_auth_level
        ))

        message = Message(
            token=NDiscountTokens.TK_NDISCOUNT_ADD_BENEFIT,
            data=data,
            data_type=DataType.string
        )

        response = self.message_bus.send_message("nDiscount", message)
        if response.token == TK_SYS_NAK:
            raise ErrorApplyingBenefit(response.data)

    def remove_benefit(self, current_pos_id, benefit_ids=None):
        # type: (int, Optional[str, list, None]) -> Message

        data = json.dumps(dict(pos_id=current_pos_id, benefit=benefit_ids))
        message = Message(token=NDiscountTokens.TK_NDISCOUNT_REMOVE,
                          data=data,
                          data_type=DataType.string)

        return self.message_bus.send_message("nDiscount", message)

    def apply_automatics_benefits(self, current_pos_id):
        # type: (str) -> Message

        message = Message(token=NDiscountTokens.TK_NDISCOUNT_APPLY_AUTOMATIC_BENEFITS,
                          data=json.dumps({"pos_id": current_pos_id}),
                          data_type=DataType.string)

        return self.message_bus.send_message("nDiscount", message)

    @abstractmethod
    def check_coupon_id(
        self,       # type: DefaultBenefitApplierRepository
        pos_id,     # type: int
        coupon_id,  # type: str
    ):
        # type: (...) -> None

        raise NotImplementedError()

    @abstractmethod
    def get_benefit_and_lock(self, coupon_id):
        # type: (str) -> str

        raise NotImplementedError()

    @abstractmethod
    def unlock_benefit(
        self,
        benefit_id,     # type: str
        pos_id,         # type: int
    ):
        # type: (...) -> None

        raise NotImplementedError()

    @abstractmethod
    def burn_benefit(
        self,
        benefit_id,     # type: str
        pos_id,         # type: int
    ):
        # type: (...) -> None

        raise NotImplementedError()

    @abstractmethod
    def get_benefit_info(self, data):
        # type: (str) -> None

        raise NotImplementedError()

    @abstractmethod
    def get_benefit_id_by_custom_property(self, order_picture):
        # type: (eTree) -> str

        raise NotImplementedError()



from __future__ import print_function

import sys
from collections import namedtuple

USING_PYTHON2 = True if sys.version_info < (3, 0) else False

def python2_input(prompt=None):
    try:
        cur_stdin, cur_stdout = sys.stdin, sys.stdout
        if hasattr(sys.stdin, '_original_stream'):
            sys.stdin = sys.stdin._original_stream
        if hasattr(sys.stdout, '_original_stream'):
            sys.stdout = sys.stdout._original_stream
        encoded_prompt = prompt.encode(getattr(sys.stdout, 'encoding', 'utf-8'))
        return raw_input(encoded_prompt).decode(getattr(sys.stdin, 'encoding', 'utf-8')) # noqa
    finally:
        sys.stdin, sys.stdout = cur_stdin, cur_stdout

Move = namedtuple('Move', ('new_name', 'old_module', 'old_name'))

class RedirectingLoader(object):
    def __init__(self, name, parent_name='eight'):
        self._name = name
        self._parent_name = parent_name
        self._moves = {}
        self._old_module = None
        self._module = None

    def _add_redirect(self, move):
        self._moves[move.new_name] = move

    def __getattr__(self, attr):
        if USING_PYTHON2 and attr in self._moves:
            move = self._moves[attr]
            if self._old_module is None:
                self._old_module = __import__(move.old_module)
            return getattr(self._old_module, move.old_name)
        else:
            if self._module is None:
                self._module = __import__(self._name)
            if len(self._moves) == 0:
                sys.modules[self._parent_name + '.' + self._name] = sys.modules[self._name]
            return getattr(self._module, attr)



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import absolute_import, division, print_function

from cryptography.hazmat.primitives.serialization.base import (
    BestAvailableEncryption, Encoding, KeySerializationEncryption,
    NoEncryption, ParameterFormat, PrivateFormat, PublicFormat,
    load_der_parameters, load_der_private_key, load_der_public_key,
    load_pem_parameters, load_pem_private_key, load_pem_public_key,
)
from cryptography.hazmat.primitives.serialization.ssh import (
    load_ssh_public_key
)


_PEM_DER = (Encoding.PEM, Encoding.DER)

__all__ = [
    "load_der_parameters", "load_der_private_key", "load_der_public_key",
    "load_pem_parameters", "load_pem_private_key", "load_pem_public_key",
    "load_ssh_public_key", "Encoding", "PrivateFormat", "PublicFormat",
    "ParameterFormat", "KeySerializationEncryption", "BestAvailableEncryption",
    "NoEncryption",
]



from ._StatisticBox import StatisticBox
from ._MeanTimeStatisticBox import MeanTimeStatisticBox
from ._OrdersCounterStatisticBox import OrdersCounterStatisticBox
from ._ReadingPercentageStatisticBox import ReadingPercentageStatisticBox
from ._TimePercentageStatisticBox import TimePercentageStatisticBox



# -*- coding: utf-8 -*-
import json
import sysactions
from typing import List

from application.interactor import NavigationInteractor
from ..models import Navigation, NavigationProduct


@sysactions.action
def get_navigation_data(_):
    # type: () -> str

    interactor = NavigationInteractor()
    navigations = _get_navigations(interactor)
    navigation_products = interactor.get_navigation_products()
    navigations = populate_navigation_products(navigations, navigation_products)

    return json.dumps(navigations, default=lambda o: getattr(o, '__dict__', str(o)))


def _get_navigations(interactor):
    # type: (NavigationInteractor) -> List[Navigation]

    navigation_list = interactor.get_navigation_list()
    main_navigations = _get_main_navigations(navigation_list)
    navigation_menus = _get_navigation_menus(main_navigations, navigation_list)

    return navigation_menus


def populate_navigation_products(navigations, navigation_products):
    # type: (List[Navigation], List[NavigationProduct]) -> List[Navigation]
    if len(navigation_products) == 0:
        return []

    for nav_menu in navigations:
        nav_menu_id = nav_menu.nav_id
        nav_menu.items.extend([prod for prod in navigation_products if prod.nav_id == nav_menu_id])
        navigation_products = [prod for prod in navigation_products if prod.nav_id != nav_menu_id]
        populate_navigation_products(nav_menu.groups, navigation_products)

    return navigations


def _get_navigation_menus(navigation_menus, navigation_list):
    # type: (List[Navigation], List[Navigation]) -> List[Navigation]
    if len(navigation_list) == 0:
        return []

    for nav_menu in navigation_menus:
        navigation_child = _get_navigations_child(nav_menu, navigation_list)
        nav_menu.groups.extend(navigation_child)
        nav_with_parents = [nav.nav_id for nav in navigation_child]
        navigation_list = [nav for nav in navigation_list if nav.nav_id not in nav_with_parents]
        _get_navigation_menus(nav_menu.groups, navigation_list)

    return navigation_menus


def _get_navigations_child(nav_menu, navigation_list):
    # type: (Navigation, List[Navigation]) -> List[Navigation]
    nav_menu_id = nav_menu.nav_id
    nav_child = [nav for nav in navigation_list if nav.parent_nav_id == nav_menu_id]

    return nav_child


def _get_main_navigations(navigation_list):
    # type: (List[Navigation]) -> List[Navigation]
    navigation_menus = []
    for nav in navigation_list:
        if not nav.parent_nav_id:
            navigation_menus.append(nav)

    for nav in navigation_menus:
        navigation_list.remove(nav)

    return navigation_menus



from basereport import SimpleReport
from report import Presenter
from typing import List

from .dto import TipReportDto
from .presenter import TipReportHeaderPresenter, TipReportBodyPresenter


class TipReportPresenter(Presenter):
    def present(self, dto):
        # type: (TipReportDto) -> SimpleReport
        presenters = [
            (TipReportHeaderPresenter(), dto.header),
            (TipReportBodyPresenter(), dto.body)
        ]  # type: List[Presenter]

        report = []
        for presenter, dto in presenters:
            report.extend(presenter.present(dto))

        return SimpleReport(report)




# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import absolute_import, division, print_function

from cryptography.exceptions import (
    UnsupportedAlgorithm, _Reasons
)
from cryptography.hazmat.backends.interfaces import HMACBackend
from cryptography.hazmat.primitives import constant_time
from cryptography.hazmat.primitives.twofactor import InvalidToken
from cryptography.hazmat.primitives.twofactor.hotp import HOTP
from cryptography.hazmat.primitives.twofactor.utils import _generate_uri


class TOTP(object):
    def __init__(self, key, length, algorithm, time_step, backend,
                 enforce_key_length=True):
        if not isinstance(backend, HMACBackend):
            raise UnsupportedAlgorithm(
                "Backend object does not implement HMACBackend.",
                _Reasons.BACKEND_MISSING_INTERFACE
            )

        self._time_step = time_step
        self._hotp = HOTP(key, length, algorithm, backend, enforce_key_length)

    def generate(self, time):
        counter = int(time / self._time_step)
        return self._hotp.generate(counter)

    def verify(self, totp, time):
        if not constant_time.bytes_eq(self.generate(time), totp):
            raise InvalidToken("Supplied TOTP value does not match.")

    def get_provisioning_uri(self, account_name, issuer):
        return _generate_uri(self._hotp, "totp", account_name, issuer, [
            ("period", int(self._time_step)),
        ])



from typing import Optional


class Configs(object):

    def __init__(
        self,
        loader_path=None,            # type: Optional[str]
        databases_path=None,         # type: Optional[str]
        backup_databases_path=None,  # type: Optional[str]
        component_name=None,         # type: Optional[str]
        required_services=None,      # type: Optional[str]
        instances=None,              # type: Optional[int]
        custom_data=None,            # type: Optional[str]
    ):
        # type: (...) -> None

        self.loader_path = loader_path
        self.databases_path = databases_path
        self.backup_databases_path = backup_databases_path
        self.component_name = component_name
        self.instances = instances
        self.required_services = required_services
        self.custom_data = custom_data
        self.alerts = None



import bustoken
import sysactions

from msgbus import TK_SYS_ACK

from .. import mb_context
from model.customexception import DeliveryCouponPrintError

TK_REMOTE_ORDER_REPRINT_DELIVERY = bustoken.create_token(bustoken.MSGPRT_LOW, "37", "216")


@sysactions.action
def print_original_delivery(pos_id, order_id):
    data = pos_id + "\0" + order_id + "\0" + ','.join(str(x) for x in list(sysactions.get_poslist()))
    msg = mb_context.MB_EasySendMessage("RemoteOrder", TK_REMOTE_ORDER_REPRINT_DELIVERY, data=data, timeout=10000000)
    if msg.token == TK_SYS_ACK:
        return
    
    raise DeliveryCouponPrintError("Could not retrieve remote order status")



# -*- coding: utf-8 -*-

from datetime import datetime

from repository import OrderRepository
from typing import List, Union
from voidedlinesreport.repository import UsersRepository

from .model import Line


class VoidedLinesReport(object):
    def __init__(self, order_repository, users_repository, store_id):
        # type: (OrderRepository, UsersRepository, str) -> None
        self.order_repository = order_repository
        self.users_repository = users_repository
        self.store_id = store_id

    def generate_voided_lines_report(self, report_pos, initial_date, end_date, only_body, operator_id):
        # type: (Union[str, None], datetime, datetime, bool, Union[int, None]) -> unicode
        voided_lines = self.get_voided_lines(report_pos, initial_date, end_date, operator_id)

        if only_body:
            return self._build_body(voided_lines).encode("utf-8") if len(voided_lines) > 0 else ""

        report = self._build_header(report_pos, initial_date, end_date)
        report += self._build_body(voided_lines) if len(voided_lines) > 0 else ""

        return report.encode("utf-8")

    def get_voided_lines(self, report_pos, initial_date, end_date, operator_id, session_id=None):
        # type: (Union[str, None], datetime, datetime, Union[int, None], Union[str, None]) -> List[Line]
        voided_lines = self.order_repository.get_voided_lines(report_pos, initial_date, end_date, operator_id, session_id)
        self.users_repository.get_operators_name(voided_lines)
        return voided_lines

    def _build_header(self, report_pos, initial_date, end_date):
        # type: (Union[str, None], datetime, datetime) -> unicode
        """
        ======================================
           Relatorio de Linhas Canceladas
         Data/hora..: 26/08/2019 15:06:26
         Periodo....: 26/08/2019 - 26/08/2019
         POS(s).....: Todos
        ======================================
        """

        pos_list = "Todos" if report_pos is None else report_pos.zfill(2)

        header = u"======================================\n"
        header += u"   Relatorio de Linhas Canceladas\n"
        header += u" Loja.......: " + self.store_id + "\n"
        header += u" Data/hora..: {}\n".format(datetime.now().strftime("%d/%m/%Y %H:%M:%S"))
        header += u" Periodo....: {} - {}\n".format(initial_date.strftime("%d/%m/%y"), end_date.strftime("%d/%m/%y"))
        header += u" POS(s).....: {}\n".format(pos_list)
        header += u"======================================\n"

        return header

    @staticmethod
    def _build_body(voided_lines):
        # type: (List[Line]) -> unicode
        """
        Operador            Qtd   Valor
             1 - 1          [  1] R$      10.9
          5813 - Desenvolve [  5] R$     77.74
        ======================================
        """
        operators_line = {}
        for line in voided_lines:
            if line.operator not in operators_line:
                operators_line[line.operator] = [line]
            else:
                operators_line[line.operator].append(line)

        body = u"Operador            Qtd   Valor\n"

        for operator in operators_line:
            operator_name = next(iter(operators_line[operator])).operator_name
            operator_qty_sum = 0
            operator_price_sum = 0.0
            for line in operators_line[operator]:
                operator_qty_sum += line.quantity
                operator_price_sum += line.price

            body += u"{0:>6} - {1:<10} [{2:>3}] R${3:>10}\n".format(str(operator)[:6],
                                                                    operator_name[:10],
                                                                    operator_qty_sum,
                                                                    operator_price_sum)

        body += u"======================================\n"
        return body



class Responses(object):

    ValidationError = 422

    BadRequestError = 400

    AuthValidation = 401

    NotFoundError = 404

    ClientCustomError = 406

    BlockedError = 409

    InUseError = 410

    NotAvailableError = 410

    OperationErrors = 500

    ExternalServicesError = 503



from typing import Optional

from application.model import Configurations
from application.service import BenefitService


class OnTableOpenInteractor(object):
    def __init__(self, configs, benefit_service):
        # type: (Configurations, BenefitService) -> None

        self.configs = configs
        self.logger = configs.logger
        self.benefit_service = benefit_service

    def execute(
            self,
            pos_id,     # type: int
            table_id    # type: str
    ):
        # type: (...) -> Optional[str]

        return self.benefit_service.apply_automatic_benefits_on_table(pos_id, table_id)



from httpmessageprocessorutil.automapped import auto_map
from model.electronic_message import ElectronicMessage


@auto_map({"electronic_message": {"name": "electronicMessage", "type": ElectronicMessage}})
class CacheRequest(object):
    def __init__(self, electronic_message):
        # type: (ElectronicMessage) -> None

        self.electronic_message = electronic_message



from typing import List  # noqa

from _CashReportBodyLineDto import CashReportBodyLineDto  # noqa


class CashReportBodyDto(object):
    def __init__(self):
        self.operator_count = 0
        self.initial_float = CashReportBodyLineDto(0, 0.0)
        self.total_sales = CashReportBodyLineDto(0, 0.0)
        self.voids = CashReportBodyLineDto(0, 0.0)
        self.net_sales = CashReportBodyLineDto(0, 0.0)
        self.tender_breakdown = []  # type: List[CashReportBodyLineDto]
        self.cash_ins = CashReportBodyLineDto(0, 0.0)
        self.cash_outs = CashReportBodyLineDto(0, 0.0)
        self.value_in_drawer = 0.0



from do_save_kiosk_customer_info import do_save_kiosk_customer_info
from do_start_kiosk_order import do_start_kiosk_order
from do_void_kiosk_order import do_void_kiosk_order
from get_kiosk_configurations import get_kiosk_configurations
from get_kiosk_prices import get_kiosk_prices



import datetime
from Queue import Queue, Empty
from logging import Logger
from threading import Thread, Lock

from production.command import ProductionCommand
from typing import Any


class UniqueQueue(Queue, object):
    def __init__(self):
        super(UniqueQueue, self).__init__()
        self._queue_lock = Lock()

    def _put(self, item):
        with self._queue_lock:
            for i in self.queue:
                if hash(i) == hash(item):
                    return

            self.queue.append(item)


class CommandQueue(object):
    def __init__(self, production_manager, logger):
        # type: (Any, Logger) -> None
        self.production_manager = production_manager
        self.logger = logger
        self.command_queue = UniqueQueue()
        self.running = True

        self.processor_thread = Thread(target=self.consume_queue)
        self.processor_thread.daemon = True
        self.processor_thread.start()

    def put(self, command):
        # type: (ProductionCommand) -> None
        self.logger.info("Command put on queue: {}".format(type(command)))
        self.command_queue.put(command)

    def consume_queue(self):
        while self.running:
            command_type = None
            try:
                self.logger.info("Command queue size: {}".format(self.command_queue.qsize()))
                command = self.command_queue.get(True, 5)
                start_time = datetime.datetime.now()
                command_type = type(command)

                self.logger.info("Starting to process command: {}".format(command_type))
                self.production_manager.process_command(command)
                end_time = datetime.datetime.now()
                time_delta = (end_time - start_time).total_seconds()
                self.logger.info("Process command time: {} - {}".format(time_delta, command_type))
            except Empty:
                continue
            except BaseException as _:  # noqa
                self.logger.exception("Error processing command: {}".format(command_type))

    def stop(self):
        self.running = False



import os
import cfgtools
import logging

from msgbus import MBEasyContext
from messagehandlerbuilder import MobilePosPrinterMessageHandlerBuilder
from helper import config_logger, import_pydevd

from mbcontextmessagehandler import MbContextMessageBus, MbContextMessageHandler


def main():
    import_pydevd(os.environ["LOADERCFG"], 9129, False)
    config_logger(os.environ["LOADERCFG"], "MobilePosPrinter")

    logger = logging.getLogger("MobilePosPrinter")

    config = cfgtools.read(os.environ["LOADERCFG"])
    pos_id = int(config.find_value("MobilePosPrinter.PosId"))

    mb_context = MBEasyContext("MobilePosPrinter")
    message_bus = MbContextMessageBus(mb_context)
    message_handler_builder = MobilePosPrinterMessageHandlerBuilder(pos_id, message_bus, logger)
    message_handler = MbContextMessageHandler(
        message_bus,
        "printer{}".format(pos_id),
        "Printer",
        "POS{}".format(pos_id),
        message_handler_builder)
    return message_handler.handle_events()



from application.repository import CurrencyConversionRepository


class CurrencyConversionRepositoryImpl(CurrencyConversionRepository):
    def get_currency_exchange(self):

        def inner_func(conn):
            rate_amount_list = conn.select(
                """
                    select RateAmount
                    from CurrencyExchange
                    where CurrencyFrom = 'BRL' and
                    CurrencyTo = 'USD' and
                    date('now') >= ValidFrom and
                    date('now') < ValidThru
                """)

            rate_amount_list = [float(x.get_entry(0)) for x in rate_amount_list]
            return rate_amount_list[0]

        return self.execute_with_connection(inner_func)



from ._CustomException import (
    ConfigurationReaderException,
    NotValidBenefit,
    ErrorLockingCoupon,
    ErrorUnlockingCoupon,
    ErrorBurningCoupon,
    ErrorObtainingCouponInfo,
    ErrorApplyingBenefit,
    BenefitIdNotFound,
    ErrorCheckingVoucher,
)



from helper import ExtendedEnum


class NotMappedEvents(ExtendedEnum):

    ORDER_MODIFIED = "ORDER_MODIFIED_"
    ORDER_MODIFIED_CLEAR_TENDERS = ORDER_MODIFIED + "CLEAR_TENDERS"
    ORDER_MODIFIED_STORED = ORDER_MODIFIED + "STORED"
    ORDER_MODIFIED_RESET_CURRENT_ORDER = ORDER_MODIFIED + "RESET_CURRENT_ORDER"
    ORDER_MODIFIED_DO_SALE = ORDER_MODIFIED + "DO_SALE"
    ORDER_MODIFIED_CREATE_ORDER = ORDER_MODIFIED + "CREATE_ORDER"
    ORDER_MODIFIED_DO_TENDER = ORDER_MODIFIED + "DO_TENDER"
    ORDER_MODIFIED_CUSTOM_PROPERTY = ORDER_MODIFIED + "CUSTOM_PROPERTY"
    ORDER_MODIFIED_ORDER_PROPERTIES_CHANGED = ORDER_MODIFIED + "ORDER_PROPERTIES_CHANGED"
    ORDER_MODIFIED_CHANGE_DIMENSION = ORDER_MODIFIED + "CHANGE_DIMENSION"
    ORDER_MODIFIED_RECALLED = ORDER_MODIFIED + "RECALLED"
    ORDER_MODIFIED_DISCOUNT_APPLIED = ORDER_MODIFIED + "DISCOUNT_APPLIED"
    ORDER_MODIFIED_DISCOUNT_REMOVED = ORDER_MODIFIED + "DISCOUNT_REMOVED"



"""
Pure-Python implementation of a Python 2-like str object for Python 3.
"""

from collections import Iterable
from numbers import Integral

from past.utils import PY2, with_metaclass


_builtin_bytes = bytes


class BaseOldStr(type):
    def __instancecheck__(cls, instance):
        return isinstance(instance, _builtin_bytes)


def unescape(s):
    """
    Interprets strings with escape sequences

    Example:
    >>> s = unescape(r'abc\\def')   # i.e. 'abc\\\\def'
    >>> print(s)
    'abc\def'
    >>> s2 = unescape('abc\\ndef')
    >>> len(s2)
    8
    >>> print(s2)
    abc
    def
    """
    return s.encode().decode('unicode_escape')
    

class oldstr(with_metaclass(BaseOldStr, _builtin_bytes)):
    """
    A forward port of the Python 2 8-bit string object to Py3
    """
    # Python 2 strings have no __iter__ method:
    @property
    def __iter__(self):
        raise AttributeError

    def __dir__(self):
        return [thing for thing in dir(_builtin_bytes) if thing != '__iter__']

    # def __new__(cls, *args, **kwargs):
    #     """
    #     From the Py3 bytes docstring:

    #     bytes(iterable_of_ints) -> bytes
    #     bytes(string, encoding[, errors]) -> bytes
    #     bytes(bytes_or_buffer) -> immutable copy of bytes_or_buffer
    #     bytes(int) -> bytes object of size given by the parameter initialized with null bytes
    #     bytes() -> empty bytes object
    #     
    #     Construct an immutable array of bytes from:
    #       - an iterable yielding integers in range(256)
    #       - a text string encoded using the specified encoding
    #       - any object implementing the buffer API.
    #       - an integer
    #     """
    #     
    #     if len(args) == 0:
    #         return super(newbytes, cls).__new__(cls)
    #     # Was: elif isinstance(args[0], newbytes):
    #     # We use type() instead of the above because we're redefining
    #     # this to be True for all unicode string subclasses. Warning:
    #     # This may render newstr un-subclassable.
    #     elif type(args[0]) == newbytes:
    #         return args[0]
    #     elif isinstance(args[0], _builtin_bytes):
    #         value = args[0]
    #     elif isinstance(args[0], unicode):
    #         if 'encoding' not in kwargs:
    #             raise TypeError('unicode string argument without an encoding')
    #         ###
    #         # Was:   value = args[0].encode(**kwargs)
    #         # Python 2.6 string encode() method doesn't take kwargs:
    #         # Use this instead:
    #         newargs = [kwargs['encoding']]
    #         if 'errors' in kwargs:
    #             newargs.append(kwargs['errors'])
    #         value = args[0].encode(*newargs)
    #         ### 
    #     elif isinstance(args[0], Iterable):
    #         if len(args[0]) == 0:
    #             # What is this?
    #             raise ValueError('unknown argument type')
    #         elif len(args[0]) > 0 and isinstance(args[0][0], Integral):
    #             # It's a list of integers
    #             value = b''.join([chr(x) for x in args[0]])
    #         else:
    #             raise ValueError('item cannot be interpreted as an integer')
    #     elif isinstance(args[0], Integral):
    #         if args[0] < 0:
    #             raise ValueError('negative count')
    #         value = b'\x00' * args[0]
    #     else:
    #         value = args[0]
    #     return super(newbytes, cls).__new__(cls, value)
        
    def __repr__(self):
        s = super(oldstr, self).__repr__()   # e.g. b'abc' on Py3, b'abc' on Py3
        return s[1:]

    def __str__(self):
        s = super(oldstr, self).__str__()   # e.g. "b'abc'" or "b'abc\\ndef'
        # TODO: fix this:
        assert s[:2] == "b'" and s[-1] == "'"
        return unescape(s[2:-1])            # e.g. 'abc'    or 'abc\ndef'

    def __getitem__(self, y):
        if isinstance(y, Integral):
            return super(oldstr, self).__getitem__(slice(y, y+1))
        else:
            return super(oldstr, self).__getitem__(y)

    def __getslice__(self, *args):
        return self.__getitem__(slice(*args))

    def __contains__(self, key):
        if isinstance(key, int):
            return False
    
    def __native__(self):
        return bytes(self)


__all__ = ['oldstr']



# -*- coding: utf-8 -*-

import logging

from mbcontextmessagehandler import MbContextMessageBus
from messagebus import Message, DataType
from msgbus import TK_STORECFG_GET, TK_SYS_NAK
from poshelper import config_logger_with_hv_logger
from typing import Optional

from application.model.customexception import ConfigurationReaderException
from application.model.configuration import Configurations
from cfgtools import read
from helper import config_logger


class Configurator(object):

    def __init__(self, loader_path):
        # type: (str) -> None

        self.loader_path = loader_path
        self.reader = read(self.loader_path)
        self.configs = Configurations()

    def read_loader_configurations(self):
        # type: () -> None

        try:
            self.configs.service_name = str(self.reader.find_value("Identification.ComponentName"))
            self.configs.persistence_name = str(self._get_config("PersistenceName") or "")
            self.configs.required_services = "|".join(self._get_config("RequiredServices", array=True) or [])
            self.configs.base_url = str(self._get_config("BaseURL") or "")
            self.configs.api_key = str(self._get_config("ApiKey") or "")
            self.configs.id_type = str(self._get_config("IdentifierType") or "DOCUMENT")

            self.configs.endpoints.retrieve_loyalty_customer_info = str(self._get_config("EndPoints.RetrieveLoyaltyId") or "")
            self.configs.endpoints.get_benefit = str(self._get_config("EndPoints.GetBenefit") or "")
            self.configs.endpoints.burn_benefit = str(self._get_config("EndPoints.BurnBenefit") or "")
            self.configs.endpoints.lock_benefit = str(self._get_config("EndPoints.LockBenefit") or "")
            self.configs.endpoints.unlock_benefit = str(self._get_config("EndPoints.ReleaseBenefit") or "")

        except Exception as ex:
            raise ConfigurationReaderException(ex)

    def configure_logger(self):
        # type: () -> None

        config_logger_with_hv_logger(
            loader_path=self.loader_path,
            log_name=self.configs.service_name,
            configure_root=True,
            root_level=logging.DEBUG,
        )
        self.configs.logger = logging.getLogger(self.configs.service_name)

    def get_store_id(self, message_bus):
        # type: (MbContextMessageBus) -> None

        self.configs.store_id = self._read_sw_config(message_bus, "Store.Id")

    @staticmethod
    def _read_sw_config(message_bus, key):
        # type: (MbContextMessageBus, str) -> Configurations

        message = Message(TK_STORECFG_GET, data_type=DataType.param, data=key)
        msg = message_bus.send_message("StoreWideConfig", message)
        if msg.token == TK_SYS_NAK:
            raise ConfigurationReaderException("Unable to get Store Wide Configuration: {}".format(key))
        return str(msg.data)

    def _get_config(self, key, array=False):
        # type: (str, Optional[bool]) -> str

        path = self.configs.service_name + "." + key
        return self.reader.find_values(path) if array else self.reader.find_value(path)



from ._LoyaltyToken import LoyaltyToken
from messagebus import TokenCreator, TokenPriority


class LoyaltyRequestToken(LoyaltyToken):
    TK_LOYALTY_GET_AND_LOCK_VOUCHER = TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group=LoyaltyToken.LOYALTY_MESSAGE_GROUP,
        number="2"
    )
    TK_LOYALTY_BURN_VOUCHER = TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group=LoyaltyToken.LOYALTY_MESSAGE_GROUP,
        number="3"
    )
    TK_LOYALTY_UNLOCK_VOUCHER = TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group=LoyaltyToken.LOYALTY_MESSAGE_GROUP,
        number="4"
    )



from typing import Any

from messageprocessorutil import UuidMessageProcessor


class NotMappedProcessor(UuidMessageProcessor):
    def __init__(self):
        # type: () -> None

        super(NotMappedProcessor, self).__init__()

    def parse_data(self, data, event_type):
        # type: (str, str) -> None

        return

    def call_business(self, param):
        # type: (Any) -> None

        return

    def format_response(self, message_bus, message, event, request, response):
        return

    def format_exception(self, message_bus, message, event, input_obj, exception):
        return

    def format_parse_exception(self, message_bus, message, event, data, exception):
        return



import json

from messagebus import DefaultToken, DataType, Message
from messageprocessorutil import UuidMessageProcessor

from application.interactor import RemoveBenefitInteractor
from application.model import Configurations
from typing import Tuple


class RemoveBenefitProcessor(UuidMessageProcessor):
    def __init__(self, configs, interactor):
        # type: (Configurations, RemoveBenefitInteractor) -> None

        super(RemoveBenefitProcessor, self).__init__()
        self.logger = configs.logger
        self.configs = configs
        self.interactor = interactor

    def parse_data(self, data, event_type):
        # type: (str, str) -> tuple

        data = json.loads(data)
        benefit_keys = data.get("benefit")
        pos_id = int(data.get("pos_id"))

        return pos_id, benefit_keys

    def call_business(self, param):
        # type: (Tuple) -> None

        pos_id, benefit_keys = param
        self.interactor.remove(pos_id, benefit_keys)

    def format_response(self, message_bus, message, event, request, response):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_ACK, DataType.string, ""))

    def format_exception(self, message_bus, message, event, input_obj, exception):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))

    def format_parse_exception(self, message_bus, message, event, data, exception):
        message_bus.reply_message(message, Message(DefaultToken.TK_SYS_NAK, DataType.param, str(exception)))



# -*- coding: utf-8 -*-

import base64
import simplejson as json

from datetime import datetime
from decimal import Decimal as D

from typing import List, Dict, Optional
from xml.etree import cElementTree as eTree

from application.configurator import Configurator
from application.model import EventData
from application.repository import MessageBusRepository


class OrderPictureInteractor(object):

    def __init__(self, configurations, repository):
        # type: (Configurator, MessageBusRepository) -> None

        self.configs = configurations
        self.repository = repository

    def build(self, event_data):
        # type: (EventData) -> str

        order_id = str(event_data.data.get("orderId"))
        session_id = event_data.data.get("sessionId")
        if session_id in [None, "None", ""]:
            order_picture = self.repository.get_order_picture(order_id)
            xml_picture = eTree.XML(order_picture)
            session_id = ""
            if xml_picture.get("sessionId") not in [None, "None", ""]:
                event_data.data = xml_picture
                session_id = xml_picture.get("sessionId")

        sale_type = int(event_data.data.get("saleType") or 0)
        payload = {
            "business-dt": self._get_business_period(event_data),
            "change": float(event_data.data.get("change") or 0),
            "creation-dttm": event_data.data.get("createdAtGMT") + "Z",
            "discount-amount": float(event_data.data.get("discountAmount") or 0),
            "discount-applied": event_data.data.get("discountsApplied"),
            "due-amount": float(event_data.data.get("dueAmount") or 0),
            "exemption": int(event_data.data.get("exemptionCode")) if event_data.data.get("exemptionCode") else None,
            "order-code": order_id,
            "order-discount-amount": float(event_data.data.get("orderDiscountAmount") or 0),
            "order-picture-custom-order-properties": self._get_custom_properties(event_data),
            "order-picture-sale-lines": self._get_sale_lines(event_data),
            "order-picture-tenders": self._get_tenders(event_data),
            "originator-code": str(int(event_data.data.get("originatorId")[-2:])),
            "pod-type": event_data.data.get("podType"),
            "pos-code": self._get_pos_code(event_data, session_id),
            "pos-user-id": self._get_pos_user_id(session_id),
            "price-basis": event_data.data.get("priceBasis"),
            "price-list": event_data.data.get("priceList"),
            "sale-type-id": sale_type if sale_type >= 0 else 0,
            "session": session_id,
            "state-id": int(event_data.data.get("stateId")),
            "store-code": str(self.configs.store_id),
            "tax-applied": float(event_data.data.get("taxTotal") or 0),
            "tax-total": float(event_data.data.get("taxTotal") or 0),
            "tip": float(event_data.data.get("tip") or 0),
            "total-after-discount": float(event_data.data.get("totalAfterDiscount") or 0),
            "total-amount": float(event_data.data.get("totalAmount") or 0),
            "total-gift": 0,
            "total-gross": float(event_data.data.get("totalGross") or 0),
            "total-tender": float(event_data.data.get("totalTender") or 0),
            "type-id": int(event_data.data.get("typeId")),
            "xml-data": self._get_fiscal_xml(event_data),
            "state-history": self._get_state_history(order_picture=event_data.data),
            "xml-data-cancel": self._get_canceled_fiscal_xml(order_picture=event_data.data),
        }

        return json.dumps(payload, use_decimal=True)

    @staticmethod
    def _get_canceled_fiscal_xml(
        order_picture,  # type: eTree.XML
    ):
        # type: (...) -> Optional[str]

        xml_canceled = order_picture.find(".//OrderProperty[@key='CANCELED_FISCAL_XML']")
        if xml_canceled is not None:
            return xml_canceled.get('value')

        return None

    def _get_fiscal_xml(self, event_data):
        # type: (EventData) -> Optional[str]

        fiscal_xml = None
        prop = event_data.data.find("CustomOrderProperties/OrderProperty[@key='FISCAL_XML']")
        if prop is not None:
            fiscal_xml = prop.get("value")
        elif self.configs.fiscal_xml:
            order_id = event_data.data.get("orderId")
            fiscal_xml = self.repository.get_fiscal_xml_from_fiscal_wrapper(order_id)

        if fiscal_xml is None:
            return None

        fiscal_xml = fiscal_xml.replace(" ", "")
        decoded_fiscal_xml = base64.b64decode(fiscal_xml)
        return base64.b64encode(decoded_fiscal_xml)

    @staticmethod
    def _get_business_period(event_data):
        # type: (EventData) -> str

        order_business_period = event_data.data.get("businessPeriod")
        return datetime.strptime(order_business_period, "%Y%m%d").strftime("%Y-%m-%d")

    @staticmethod
    def _get_pos_user_id(session_id):
        # type: (str) -> str

        if session_id == "":
            return '-1'

        for session_data in session_id.split(","):
            key, value = session_data.split("=")
            if key == "user":
                return value

    @staticmethod
    def _get_pos_code(event_data, session_id):
        # type: (EventData, str) -> str

        pos_id = event_data.data.get("posId")
        if pos_id is None and session_id != "":
            for session_data in session_id.split(","):
                key, value = session_data.split("=")
                if key == "pos":
                    pos_id = value
                    break

        if pos_id is None:
            pos_id = event_data.data.get("instance")

        return pos_id

    def _get_custom_properties(self, event_data):
        # type: (EventData) -> Dict

        payload_custom_order_properties = {}
        store_division = str(self.configs.store_division) if self.configs.store_division else None
        if store_division:
            payload_custom_order_properties["store-division"] = store_division

        custom_properties = event_data.data.find("CustomOrderProperties") or []
        for custom_property in custom_properties:
            if custom_property.get("key").upper() != "FISCAL_XML":
                custom_property_key = custom_property.get("value").encode("utf-8")
                payload_custom_order_properties[custom_property.get("key")] = custom_property_key

        return payload_custom_order_properties

    def _get_tenders(self, event_data):
        # type: (EventData) -> List

        payload_order_tenders = []
        for tender_line in event_data.data.findall("TenderHistory/Tender"):
            tender_payload = {
                "change-amount": float(tender_line.get("change") or 0),
                "order-picture-id": int(event_data.data.get("orderId")),
                "reference-amount": float(tender_line.get("tenderAmount") or 0),
                "tender-amount": float(tender_line.get("tenderAmount") or 0),
                "tender-detail": self._get_tender_detail(tender_line),
                "tender-type": tender_line.get("tenderType"),
                "tip": float(tender_line.get("tip")) if tender_line.get("tip") else None
            }
            payload_order_tenders.append(tender_payload)

        return payload_order_tenders

    def _get_tender_detail(self, tender_line):
        tender_detail_json = tender_line.get("tenderDetail")
        if not tender_detail_json:
            return None

        try:
            tender_detail = json.loads(tender_detail_json)
            payment_brand = tender_detail.get("Bandeira")
        except (json.JSONDecodeError, TypeError):
            return None

        if not payment_brand:
            return None

        self._fix_missing_tender_details(tender_detail)
        return json.dumps(tender_detail)

    @staticmethod
    def _fix_missing_tender_details(tender_detail):
        for value in tender_detail.keys():
            if tender_detail[value] is None:
                tender_detail[value] = ""

    def _get_sale_lines(self, event_data):
        # type: (EventData) -> List

        payload_order_sale_lines = []
        for sale_line in event_data.data.findall("SaleLine"):
            sale_line_payload = {
                "added-qty": D(sale_line.get("addedQty")) if sale_line.get("addedQty") else None,
                "added-unit-price": float(sale_line.get("addedUnitPrice") or 0),
                "chosen-qty": D(sale_line.get("chosenQty")) if sale_line.get("chosenQty") else None,
                "comment": self._get_sale_line_comments(sale_line),
                "custom-properties": self._get_sale_line_custom_properties(sale_line),
                "default-qty": D(sale_line.get("defaultQty")) if sale_line.get("defaultQty") else None,
                "inc-qty": D(sale_line.get("incQty") or 0),
                "dec-qty": D(sale_line.get("decQty") or 0),
                "item-discount": float(sale_line.get("itemDiscount") or 0),
                "discount-applied": sale_line.get("discountsApplied"),
                "item-price": float(sale_line.get("itemPrice") or 0),
                "item-type": sale_line.get("itemType"),
                "line-number": int(sale_line.get("lineNumber")),
                "level": int(sale_line.get("level")),
                "multiplied-qty": D(sale_line.get("multipliedQty") or sale_line.get("qty") or 0),
                "qty": D(sale_line.get("qty") or 0),
                "menu-item-code": sale_line.get("itemId"),
                "price-key": sale_line.get("priceKey"),
                "product": sale_line.get("productName").encode("utf-8"),
                "part-code": sale_line.get("partCode"),
                "plu": sale_line.get("partCode"),
                "order-picture-id": int(event_data.data.get("orderId")),
                "sub-qty": D(sale_line.get("subQty")) if sale_line.get("subQty") else None,
                "sub-unit-price": float(sale_line.get("subUnitPrice") or 0),
                "tax-items": self._get_sale_line_tax_items(sale_line),
                "tags": self._get_sale_line_tags(sale_line),
                "unit-price": float(sale_line.get("unitPrice") or 0)
            }
            payload_order_sale_lines.append(sale_line_payload)

        return payload_order_sale_lines

    @staticmethod
    def _get_state_history(
        order_picture,  # type: eTree
    ):
        # type: (...) -> List[dict]

        list_state_history = []
        for state in order_picture.find('StateHistory').findall('State'):
            list_state_history.append({
                'state': state.get('state'),
                'state-id': state.get('stateId'),
                'timestamp': state.get('timestamp'),
                'timestamp-gmt': state.get('timestampGMT'),
            })
        return list_state_history

    @staticmethod
    def _get_sale_line_comments(sale_line):
        # type: (eTree) -> List

        comments = []
        for comment in sale_line.findall("Comment"):
            comments.append(comment.get("comment").encode("utf-8"))

        return comments

    @staticmethod
    def _get_sale_line_custom_properties(sale_line):
        # type: (eTree) -> Dict

        custom_properties = {}
        try:
            custom_properties = json.loads(sale_line.get("customProperties") or {})
        except (json.JSONDecodeError, TypeError):
            pass
        return custom_properties

    @staticmethod
    def _get_sale_line_tax_items(sale_line):
        # type: (eTree) -> List

        tax_items = []
        for order_tax_item in sale_line.findall("TaxItem"):
            tax_item = {
                "base-amount-ad": float(order_tax_item.get("baseAmountAD", 0)),
                "base-amount-bd": float(order_tax_item.get("baseAmountBD", 0)),
                "tax-amount-ad": float(order_tax_item.get("taxAmountAD", 0)),
                "tax-amount-bd": float(order_tax_item.get("taxAmountBD", 0)),
                "tax-index": order_tax_item.get("taxIndex"),
                "tax-rate": float(order_tax_item.get("taxRate", 0)),
                "tax-rule-id": int(order_tax_item.get("taxRuleId"))
            }
            tax_items.append(tax_item)

        return tax_items

    @staticmethod
    def _get_sale_line_tags(sale_line):
        # type: (eTree) -> List

        item_tags = []
        try:
            json_array_tags = json.loads(sale_line.get("jsonArrayTags") or "[]")
            for temp_tag in json_array_tags:
                item_tags.append(temp_tag)
        except (json.JSONDecodeError, TypeError):
            pass

        return item_tags



from production.box._OrderChangerProductionBox import OrderChangerProductionBox


class SendMomentTimerBox(OrderChangerProductionBox):
    def change_order(self, order):
        send_moment_list = []
        self.find_recursive_send_moment(order, send_moment_list)

        if len(send_moment_list) > 0:
            min_send_moment = min(send_moment_list)
            order.display_time = min_send_moment.isoformat()

        return order

    def find_recursive_send_moment(self, order, send_moment_list):
        for item in order.items:
            if item.is_product():
                if not item.has_tag("dont-need-cook") and not item.has_tag("wait-prep-time") and item.send_moment:
                    send_moment_list.append(item.send_moment)
            elif len(item.items) > 0:
                self.find_recursive_send_moment(item, send_moment_list)



from logging import Logger

import requests
from helper import json_serialize, retry
from model.config import Config
from model.electronic_message import ElectronicMessage
from model.error_type import ExternalServicesErrorsType
from model.exception import ExternalServiceErrors
from model.services import ApiServiceModel


class ApiService(ApiServiceModel):

    def __init__(self, config, logger):
        # type: (Config, Logger) -> None

        self.config = config
        self.logger = logger

    def send_electronic_messages(self, electronic_messages):
        # type: (ElectronicMessage) -> None

        url = "http://{}:{}/cache/peripheral/electronicPayment".format(self.config.host, self.config.port)
        headers = {"Content-type": "application/json"}
        data = json_serialize({"electronicMessage": electronic_messages})

        response = retry(
                logger=self.logger,
                method_name="send_electronic_messages",
                n=3,
                sleep_seconds=5,
                trying_to_do=lambda: requests.request("POST", url=url, headers=headers, timeout=5, data=data),
                is_success=lambda resp: resp.status_code == 204,
                success=lambda resp: resp,
                failed=lambda resp: resp)

        if not response:
            raise ExternalServiceErrors(ExternalServicesErrorsType.ElectronicMessageUnavailable)



#!/usr/bin/env python
#
# This is a module that gathers a list of serial ports on POSIXy systems.
# For some specific implementations, see also list_ports_linux, list_ports_osx
#
# This file is part of pySerial. https://github.com/pyserial/pyserial
# (C) 2011-2015 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause

"""\
The ``comports`` function is expected to return an iterable that yields tuples
of 3 strings: port name, human readable description and a hardware ID.

As currently no method is known to get the second two strings easily, they are
currently just identical to the port name.
"""

from __future__ import absolute_import

import glob
import sys
import os
from serial.tools import list_ports_common

# try to detect the OS so that a device can be selected...
plat = sys.platform.lower()

if plat[:5] == 'linux':    # Linux (confirmed)  # noqa
    from serial.tools.list_ports_linux import comports

elif plat[:6] == 'darwin':   # OS X (confirmed)
    from serial.tools.list_ports_osx import comports

elif plat == 'cygwin':       # cygwin/win32
    # cygwin accepts /dev/com* in many contexts
    # (such as 'open' call, explicit 'ls'), but 'glob.glob'
    # and bare 'ls' do not; so use /dev/ttyS* instead
    def comports(include_links=False):
        devices = glob.glob('/dev/ttyS*')
        if include_links:
            devices.extend(list_ports_common.list_links(devices))
        return [list_ports_common.ListPortInfo(d) for d in devices]

elif plat[:7] == 'openbsd':    # OpenBSD
    def comports(include_links=False):
        devices = glob.glob('/dev/cua*')
        if include_links:
            devices.extend(list_ports_common.list_links(devices))
        return [list_ports_common.ListPortInfo(d) for d in devices]

elif plat[:3] == 'bsd' or plat[:7] == 'freebsd':
    def comports(include_links=False):
        devices = glob.glob('/dev/cua*[!.init][!.lock]')
        if include_links:
            devices.extend(list_ports_common.list_links(devices))
        return [list_ports_common.ListPortInfo(d) for d in devices]

elif plat[:6] == 'netbsd':   # NetBSD
    def comports(include_links=False):
        """scan for available ports. return a list of device names."""
        devices = glob.glob('/dev/dty*')
        if include_links:
            devices.extend(list_ports_common.list_links(devices))
        return [list_ports_common.ListPortInfo(d) for d in devices]

elif plat[:4] == 'irix':     # IRIX
    def comports(include_links=False):
        """scan for available ports. return a list of device names."""
        devices = glob.glob('/dev/ttyf*')
        if include_links:
            devices.extend(list_ports_common.list_links(devices))
        return [list_ports_common.ListPortInfo(d) for d in devices]

elif plat[:2] == 'hp':       # HP-UX (not tested)
    def comports(include_links=False):
        """scan for available ports. return a list of device names."""
        devices = glob.glob('/dev/tty*p0')
        if include_links:
            devices.extend(list_ports_common.list_links(devices))
        return [list_ports_common.ListPortInfo(d) for d in devices]

elif plat[:5] == 'sunos':    # Solaris/SunOS
    def comports(include_links=False):
        """scan for available ports. return a list of device names."""
        devices = glob.glob('/dev/tty*c')
        if include_links:
            devices.extend(list_ports_common.list_links(devices))
        return [list_ports_common.ListPortInfo(d) for d in devices]

elif plat[:3] == 'aix':      # AIX
    def comports(include_links=False):
        """scan for available ports. return a list of device names."""
        devices = glob.glob('/dev/tty*')
        if include_links:
            devices.extend(list_ports_common.list_links(devices))
        return [list_ports_common.ListPortInfo(d) for d in devices]

else:
    # platform detection has failed...
    import serial
    sys.stderr.write("""\
don't know how to enumerate ttys on this system.
! I you know how the serial ports are named send this information to
! the author of this module:

sys.platform = {!r}
os.name = {!r}
pySerial version = {}

also add the naming scheme of the serial ports and with a bit luck you can get
this module running...
""".format(sys.platform, os.name, serial.VERSION))
    raise ImportError("Sorry: no implementation for your platform ('{}') available".format(os.name))

# test
if __name__ == '__main__':
    for port, desc, hwid in sorted(comports()):
        print("{}: {} [{}]".format(port, desc, hwid))



from _SatXmlPartBuilder import SatXmlPartBuilder
from _CfeBuilder import CfeBuilder
from _ContextKeys import ContextKeys
from _DestinatarioBuilder import DestinatarioBuilder
from _DetListBuilder import DetListBuilder
from _DetBuilder import DetBuilder
from _DetTipBuilder import DetTipBuilder
from _EmitBuilder import EmitBuilder
from _IdeBuilder import IdeBuilder
from _InfCfeBuilder import InfCfeBuilder
from _PgtoBuilder import PgtoBuilder
from _TotalBuilder import TotalBuilder
from _TaxBuilder import TaxBuilder, TaxBuilderLucroReal, TaxBuilderSimplesNacional
from _InfAdicBuilder import InfAdicBuilder

__all__ = [ "CfeBuilder",
            "ContextKeys",
            "DestinatarioBuilder",
            "DetListBuilder",
            "DetBuilder",
            "DetTipBuilder",
            "EmitBuilder",
            "IdeBuilder",
            "InfCfeBuilder",
            "PgtoBuilder",
            "TotalBuilder",
            "TaxBuilder",
            "TaxBuilderLucroReal",
            "TaxBuilderSimplesNacional",
            "InfAdicBuilder"
           ]



from xml.etree import cElementTree as eTree

from application.interactor import UnlockBenefitInteractor
from application.model import BenefitAppliers
from application.model.configuration import Configurations
from application.model.repository.benefit_applier import DefaultBenefitApplierRepository
from application.repository.pos import DBRepository
from application.repository.pos.utils import find_order_benefits_id
from typing import Dict


class OrderModifiedVoidInteractor(object):

    def __init__(self, unlock_benefit_interactor, configs, db_repository, benefit_appliers_repositories):
        # type: (UnlockBenefitInteractor, Configurations, DBRepository, Dict[BenefitAppliers: DefaultBenefitApplierRepository]) -> None

        self.configs = configs
        self.logger = self.configs.logger
        self.db_repository = db_repository
        self.benefit_appliers_repositories = benefit_appliers_repositories
        self.unlock_benefit_interactor = unlock_benefit_interactor

    def unlock_voided_benefit(self, current_pos_id, originator_pos_id, order_id, order_picture):
        # type: (int, int, int, eTree) -> None

        order_benefits_id = find_order_benefits_id(self.logger, order_picture)
        if not order_benefits_id:
            return

        message = "Benefits founded to unlock: {}; Current/Originator PosId: {}/{}; OrderId: {}"
        self.logger.info(message.format(order_benefits_id, current_pos_id, originator_pos_id, order_id))

        for benefit_id in order_benefits_id:
            benefit_applier = self.db_repository.get_benefit_applier_by_db(benefit_id, originator_pos_id)
            if benefit_applier not in [None, 'None', '']:
                self.unlock_benefit_interactor.unlock_benefit(benefit_id, originator_pos_id, benefit_applier)



# -*- coding: utf-8 -*-
import sysactions
from actions.login.common_close_user import set_authorizer
from typing import Optional

from actions.login.signout_user import sign_out_user
from actions.util.list_users import get_opened_users
from custom_sysactions import user_authorized_to_execute_action


@sysactions.action
def close_operators_manager_menu(
    pos_id,                         # type: str
    current_operator_id=None,       # type: Optional[str]
):
    # type: (...) -> bool

    model = sysactions.get_model(pos_id)
    manager_id = sysactions.get_custom(
        model=model,
        key='Last Manager ID',
    )

    logged_out_user = execute_sign_out_user(
        pos_id=pos_id,
        current_operator_id=current_operator_id,
        auth_data=manager_id,
    )

    set_authorizer(
        pos_id=pos_id,
        authorizer=manager_id,
        is_quick_service=False,
    )

    return logged_out_user


@sysactions.action
@user_authorized_to_execute_action
def close_operators(
    pos_id,                  # type: str
    current_operator_id,     # type: str
    auth_data=None,          # type: Optional[str]
):
    # type: (...) -> bool

    return execute_sign_out_user(
        pos_id=pos_id,
        current_operator_id=current_operator_id,
        auth_data=auth_data,
    )


def execute_sign_out_user(
    pos_id,                 # type: str
    current_operator_id,    # type: str
    auth_data=None,         # type: Optional[str]
):
    # type: (...) -> bool

    model = sysactions.get_model(posid=pos_id)
    sysactions.check_current_order(
        posid=pos_id,
        model=model,
        need_order=False,
    )

    pos_to_close_user, user, _ = get_opened_users(pos_id=pos_id)

    if user is None:
        return False

    return sign_out_user(
        pos_id=pos_id,
        user_id=user.get("id"),
        pos_to_close_user=pos_to_close_user,
        current_operator_id=current_operator_id,
        auth_data=auth_data,
    )



class Navigation(object):
    def __init__(self, nav_id, name, parent_nav_id, sort_order, button_text, background_color, restricted_period):
        self.nav_id = nav_id
        self.name = name
        self.parent_nav_id = parent_nav_id
        self.sort_order = sort_order
        self.button_text = button_text
        self.background_color = background_color
        self.groups = []
        self.items = []
        self.restricted_service_period = [] if restricted_period is None else restricted_period.split(",")



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK
""" CLI

This module acts as a command line interface for python-escpos. It mirrors
closely the available ESCPOS commands while adding a couple extra ones for convience.

It requires you to have a configuration file. See documentation for details.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import argparse
try:
    import argcomplete
except ImportError:
    # this CLI works nevertheless without argcomplete
    pass  # noqa
import sys
import six
from . import config
from . import version


# Must be defined before it's used in DEMO_FUNCTIONS
def str_to_bool(string):
    """ Used as a type in argparse so that we get back a proper
    bool instead of always True
    """
    return string.lower() in ('y', 'yes', '1', 'true')

# A list of functions that work better with a newline to be sent after them.
REQUIRES_NEWLINE = ('qr', 'barcode', 'text', 'block_text')

# Used in demo method
# Key: The name of escpos function and the argument passed on the CLI. Some
#   manual translation is done in the case of barcodes_a -> barcode.
# Value: A list of dictionaries to pass to the escpos function as arguments.
DEMO_FUNCTIONS = {
    'text': [
        {'txt': 'Hello, World!\n', }
    ],
    'qr': [
        {'content': 'This tests a QR code'},
        {'content': 'https://en.wikipedia.org/'}
    ],
    'barcodes_a': [
        {'bc': 'UPC-A', 'code': '13243546576'},
        {'bc': 'UPC-E', 'code': '132435'},
        {'bc': 'EAN13', 'code': '1324354657687'},
        {'bc': 'EAN8', 'code': '1324354'},
        {'bc': 'CODE39', 'code': 'TEST'},
        {'bc': 'ITF', 'code': '55867492279103'},
        {'bc': 'NW7', 'code': 'A00000000A'},
    ],
    'barcodes_b': [
        {'bc': 'UPC-A', 'code': '13243546576', 'function_type': 'B'},
        {'bc': 'UPC-E', 'code': '132435', 'function_type': 'B'},
        {'bc': 'EAN13', 'code': '1324354657687', 'function_type': 'B'},
        {'bc': 'EAN8', 'code': '1324354', 'function_type': 'B'},
        {'bc': 'CODE39', 'code': 'TEST', 'function_type': 'B'},
        {'bc': 'ITF', 'code': '55867492279103', 'function_type': 'B'},
        {'bc': 'NW7', 'code': 'A00000000A', 'function_type': 'B'},
        {'bc': 'CODE93', 'code': 'A00000000A', 'function_type': 'B'},
        {'bc': 'CODE93', 'code': '1324354657687', 'function_type': 'B'},
        {'bc': 'CODE128A', 'code': 'TEST', 'function_type': 'B'},
        {'bc': 'CODE128B', 'code': 'TEST', 'function_type': 'B'},
        {'bc': 'CODE128C', 'code': 'TEST', 'function_type': 'B'},
        {'bc': 'GS1-128', 'code': '00123456780000000001', 'function_type': 'B'},
        {'bc': 'GS1 DataBar Omnidirectional', 'code': '0000000000000', 'function_type': 'B'},
        {'bc': 'GS1 DataBar Truncated', 'code': '0000000000000', 'function_type': 'B'},
        {'bc': 'GS1 DataBar Limited', 'code': '0000000000000', 'function_type': 'B'},
        {'bc': 'GS1 DataBar Expanded', 'code': '00AAAAAAA', 'function_type': 'B'},
    ],
}

# Used to build the CLI
# A list of dictionaries. Each dict is a CLI argument.
# Keys:
# parser: A dict of args for command_parsers.add_parser
# defaults: A dict of args for subparser.set_defaults
# arguments: A list of dicts of args for subparser.add_argument
ESCPOS_COMMANDS = [
    {
        'parser': {
            'name': 'qr',
            'help': 'Print a QR code',
        },
        'defaults': {
            'func': 'qr',
        },
        'arguments': [
            {
                'option_strings': ('--content',),
                'help': 'Text to print as a qr code',
                'required': True,
            }
        ],
    },
    {
        'parser': {
            'name': 'barcode',
            'help': 'Print a barcode',
        },
        'defaults': {
            'func': 'barcode',
        },
        'arguments': [
            {
                'option_strings': ('--code',),
                'help': 'Barcode data to print',
                'required': True,
            },
            {
                'option_strings': ('--bc',),
                'help': 'Barcode format',
                'required': True,
            },
            {
                'option_strings': ('--height',),
                'help': 'Barcode height in px',
                'type': int,
            },
            {
                'option_strings': ('--width',),
                'help': 'Barcode width',
                'type': int,
            },
            {
                'option_strings': ('--pos',),
                'help': 'Label position',
                'choices': ['BELOW', 'ABOVE', 'BOTH', 'OFF'],
            },
            {
                'option_strings': ('--font',),
                'help': 'Label font',
                'choices': ['A', 'B'],
            },
            {
                'option_strings': ('--align_ct',),
                'help': 'Align barcode center',
                'type': str_to_bool,
            },
            {
                'option_strings': ('--function_type',),
                'help': 'ESCPOS function type',
                'choices': ['A', 'B'],
            },
        ],
    },
    {
        'parser': {
            'name': 'text',
            'help': 'Print plain text',
        },
        'defaults': {
            'func': 'text',
        },
        'arguments': [
            {
                'option_strings': ('--txt',),
                'help': 'Plain text to print',
                'required': True,
            }
        ],
    },
    {
        'parser': {
            'name': 'block_text',
            'help': 'Print wrapped text',
        },
        'defaults': {
            'func': 'block_text',
        },
        'arguments': [
            {
                'option_strings': ('--txt',),
                'help': 'block_text to print',
                'required': True,
            },
            {
                'option_strings': ('--columns',),
                'help': 'Number of columns',
                'type': int,
            },
        ],
    },
    {
        'parser': {
            'name': 'cut',
            'help': 'Cut the paper',
        },
        'defaults': {
            'func': 'cut',
        },
        'arguments': [
            {
                'option_strings': ('--mode',),
                'help': 'Type of cut',
                'choices': ['FULL', 'PART'],
            },
        ],
    },
    {
        'parser': {
            'name': 'cashdraw',
            'help': 'Kick the cash drawer',
        },
        'defaults': {
            'func': 'cashdraw',
        },
        'arguments': [
            {
                'option_strings': ('--pin',),
                'help': 'Which PIN to kick',
                'choices': [2, 5],
            },
        ],
    },
    {
        'parser': {
            'name': 'image',
            'help': 'Print an image',
        },
        'defaults': {
            'func': 'image',
        },
        'arguments': [
            {
                'option_strings': ('--img_source',),
                'help': 'Path to image',
                'required': True,
            },
            {
                'option_strings': ('--impl',),
                'help': 'Implementation to use',
                'choices': ['bitImageRaster', 'bitImageColumn', 'graphics'],
            },
            {
                'option_strings': ('--high_density_horizontal',),
                'help': 'Image density (horizontal)',
                'type': str_to_bool,
            },     
            {
                'option_strings': ('--high_density_vertical',),
                'help': 'Image density (vertical)',
                'type': str_to_bool,
            },      
                      
        ],
    },
    {
        'parser': {
            'name': 'fullimage',
            'help': 'Print a fullimage',
        },
        'defaults': {
            'func': 'fullimage',
        },
        'arguments': [
            {
                'option_strings': ('--img',),
                'help': 'Path to img',
                'required': True,
            },
            {
                'option_strings': ('--max_height',),
                'help': 'Max height of image in px',
                'type': int,
            },
            {
                'option_strings': ('--width',),
                'help': 'Max width of image in px',
                'type': int,
            },
            {
                'option_strings': ('--histeq',),
                'help': 'Equalize the histrogram',
                'type': str_to_bool,
            },
            {
                'option_strings': ('--bandsize',),
                'help': 'Size of bands to divide into when printing',
                'type': int,
            },
        ],
    },
    {
        'parser': {
            'name': 'charcode',
            'help': 'Set character code table',
        },
        'defaults': {
            'func': 'charcode',
        },
        'arguments': [
            {
                'option_strings': ('--code',),
                'help': 'Character code',
                'required': True,
            },
        ],
    },
    {
        'parser': {
            'name': 'set',
            'help': 'Set text properties',
        },
        'defaults': {
            'func': 'set',
        },
        'arguments': [
            {
                'option_strings': ('--align',),
                'help': 'Horizontal alignment',
                'choices': ['left', 'center', 'right'],
            },
            {
                'option_strings': ('--font',),
                'help': 'Font choice',
                'choices': ['left', 'center', 'right'],
            },
            {
                'option_strings': ('--text_type',),
                'help': 'Text properties',
                'choices': ['B', 'U', 'U2', 'BU', 'BU2', 'NORMAL'],
            },
            {
                'option_strings': ('--width',),
                'help': 'Width multiplier',
                'type': int,
            },
            {
                'option_strings': ('--height',),
                'help': 'Height multiplier',
                'type': int,
            },
            {
                'option_strings': ('--density',),
                'help': 'Print density',
                'type': int,
            },
            {
                'option_strings': ('--invert',),
                'help': 'White on black printing',
                'type': str_to_bool,
            },
            {
                'option_strings': ('--smooth',),
                'help': 'Text smoothing. Effective on >:  4x4 text',
                'type': str_to_bool,
            },
            {
                'option_strings': ('--flip',),
                'help': 'Text smoothing. Effective on >:  4x4 text',
                'type': str_to_bool,
            },
        ],
    },
    {
        'parser': {
            'name': 'hw',
            'help': 'Hardware operations',
        },
        'defaults': {
            'func': 'hw',
        },
        'arguments': [
            {
                'option_strings': ('--hw',),
                'help': 'Operation',
                'choices': ['INIT', 'SELECT', 'RESET'],
                'required': True,
            },
        ],
    },
    {
        'parser': {
            'name': 'control',
            'help': 'Control sequences',
        },
        'defaults': {
            'func': 'control',
        },
        'arguments': [
            {
                'option_strings': ('--ctl',),
                'help': 'Control sequence',
                'choices': ['LF', 'FF', 'CR', 'HT', 'VT'],
                'required': True,
            },
            {
                'option_strings': ('--pos',),
                'help': 'Horizontal tab position (1-4)',
                'type': int,
            },
        ],
    },
    {
        'parser': {
            'name': 'panel_buttons',
            'help': 'Controls panel buttons',
        },
        'defaults': {
            'func': 'panel_buttons',
        },
        'arguments': [
            {
                'option_strings': ('--enable',),
                'help': 'Feed button enabled',
                'type': str_to_bool,
                'required': True,
            },
        ],
    },
    {
        'parser': {
            'name': 'raw',
            'help': 'Raw data',
        },
        'defaults': {
            'func': '_raw',
        },
        'arguments': [
            {
                'option_strings': ('--msg',),
                'help': 'Raw data to send',
                'required': True,
            },
        ],
    },
]


def main():
    """

    Handles loading of configuration and creating and processing of command
    line arguments. Called when run from a CLI.

    """

    parser = argparse.ArgumentParser(
        description='CLI for python-escpos',
        epilog='Printer configuration is defined in the python-escpos config'
        'file. See documentation for details.',
    )

    parser.register('type', 'bool', str_to_bool)

    # Allow config file location to be passed
    parser.add_argument(
        '-c', '--config',
        help='Alternate path to the configuration file',
    )

    # Everything interesting runs off of a subparser so we can use the format
    # cli [subparser] -args
    command_subparsers = parser.add_subparsers(
        title='ESCPOS Command',
        dest='parser',
    )
    # fix inconsistencies in the behaviour of some versions of argparse
    command_subparsers.required = False   # force 'required' testing

    # Build the ESCPOS command arguments
    for command in ESCPOS_COMMANDS:
        parser_command = command_subparsers.add_parser(**command['parser'])
        parser_command.set_defaults(**command['defaults'])
        for argument in command['arguments']:
            option_strings = argument.pop('option_strings')
            parser_command.add_argument(*option_strings, **argument)

    # Build any custom arguments
    parser_command_demo = command_subparsers.add_parser('demo',
                                                        help='Demonstrates various functions')
    parser_command_demo.set_defaults(func='demo')
    demo_group = parser_command_demo.add_mutually_exclusive_group()
    demo_group.add_argument(
        '--barcodes-a',
        help='Print demo barcodes for function type A',
        action='store_true',
    )
    demo_group.add_argument(
        '--barcodes-b',
        help='Print demo barcodes for function type B',
        action='store_true',
    )
    demo_group.add_argument(
        '--qr',
        help='Print some demo QR codes',
        action='store_true',
    )
    demo_group.add_argument(
        '--text',
        help='Print some demo text',
        action='store_true',
    )

    parser_command_version = command_subparsers.add_parser('version',
                                                           help='Print the version of python-escpos')
    parser_command_version.set_defaults(version=True)

    # hook in argcomplete
    if 'argcomplete' in globals():
        argcomplete.autocomplete(parser)

    # Get only arguments actually passed
    args_dict = vars(parser.parse_args())
    if not args_dict:
        parser.print_help()
        sys.exit()
    command_arguments = dict([k, v] for k, v in six.iteritems(args_dict) if v is not None)

    # If version should be printed, do this, then exit
    print_version = command_arguments.pop('version', None)
    if print_version:
        print(version.version)
        sys.exit()

    # If there was a config path passed, grab it
    config_path = command_arguments.pop('config', None)

    # Load the configuration and defined printer
    saved_config = config.Config()
    saved_config.load(config_path)
    printer = saved_config.printer()

    if not printer:
        raise Exception('No printers loaded from config')

    target_command = command_arguments.pop('func')

    # remove helper-argument 'parser' from dict
    command_arguments.pop('parser', None)

    if hasattr(printer, target_command):
        # print command with args
        getattr(printer, target_command)(**command_arguments)
        if target_command in REQUIRES_NEWLINE:
            printer.text("\n")
    else:
        command_arguments['printer'] = printer
        globals()[target_command](**command_arguments)


def demo(printer, **kwargs):
    """
    Prints specificed demos. Called when CLI is passed `demo`. This function
    uses the DEMO_FUNCTIONS dictionary.

    :param printer: A printer from escpos.printer
    :param kwargs: A dict with a key for each function you want to test. It's
        in this format since it usually comes from argparse.
    """
    for demo_choice in kwargs.keys():
        command = getattr(
            printer,
            demo_choice
            .replace('barcodes_a', 'barcode')
            .replace('barcodes_b', 'barcode')
        )
        for params in DEMO_FUNCTIONS[demo_choice]:
            command(**params)
        printer.cut()

if __name__ == '__main__':
    main()



import json
import logging
import sysactions

from bustoken import TK_ELECTRONIC_PAYMENT_CACHE
from msgbus import (
    FM_PARAM,
    TK_SYS_ACK,
    MBMessage,
)
from sysactions import mbcontext
from typing import (
    Dict,
    Optional,
    List,
)

from actions.eft import get_electronic_payment_component_name
from tef import PrintReceipt

logger = logging.getLogger("PosActions")


def set_eft_cache(
    pos_id,             # type: str
    payment_data,       # type: Dict
    tender_id,          # type: int
    order_id,           # type: int
):
    # type: (...) -> None

    try:
        receipt_customer = payment_data.get("ReceiptCustomer", None)
        if receipt_customer is None:
            return

        receipt_data = {
            "order_id": order_id,
            "tender_id": tender_id,
            "receipt_customer": receipt_customer,
        }
        receipt_data = json.dumps(receipt_data)

        data = json.dumps(
            {
                "posId": pos_id,
                "eftProcessor": payment_data.get("TransactionProcessor"),
                "receiptCustomer": receipt_data,
             }
        )
        response = _send_message(
            pos_id=pos_id,
            data=data,
        )
        if response.token != TK_SYS_ACK:
            message = "NACK reprint last eft receipt. {} - {} - {}"
            logger.info(message.format(pos_id, response.token, response.data))

    except Exception as _:
        logger.info("Error set last eft receipt")


@sysactions.action
def reprint_last_eft_receipt(
    pos_id,                     # type: str
    eft_processor="",           # type: Optional[str]
    fiscalization_type="",      # type: Optional[str]
    error_title="$EFT",         # type: Optional[str]
):
    # type: (...) -> Dict

    error_message = "$ERROR_REPRINT_TEF_RECEIPT"
    eft_check = _check_eft_value(
        pos_id=pos_id,
        eft_value=eft_processor,
        error_message=error_message,
        error_title=error_title,
    )
    if not eft_check:
        return {}

    try:
        data = json.dumps(
            {
                "posId": pos_id,
                "eftProcessor": eft_processor,
            }
        )
        response = _send_message(
            pos_id=pos_id,
            data=data,
        )
        if response.token != TK_SYS_ACK or response.data in ['', None]:
            message = "NACK reprint last eft receipt. {} - {} - {}"
            logger.info(message.format(pos_id, response.token, response.data))
            sysactions.show_messagebox(
                posid=pos_id,
                message=error_message,
                title="$EFT",
                buttons="$OK",
            )
            return {}

        response_dict = json.loads(response.data)
        if response_dict.get('receipt_customer', {}) in [None, '']:
            return response_dict

        logger.info("Success reprint last eft receipt. {}".format(pos_id))
        reprint_receipt(
            pos_id=pos_id,
            receipts=[response_dict.get('receipt_customer')],
            fiscalization_type=fiscalization_type,
            error_title=error_title,
        )
        return response_dict
    except Exception as _:
        logger.info("Error reprint last eft receipt")


def _send_message(
    pos_id,     # type: str
    data,       # type: str
):
    # type: (...) -> MBMessage

    sitef_comp_name = get_electronic_payment_component_name(pos_id=pos_id)
    return mbcontext.MB_EasySendMessage(
        dest_name=sitef_comp_name,
        token=TK_ELECTRONIC_PAYMENT_CACHE,
        format=FM_PARAM,
        data=data,
        timeout=600000000,
    )


def reprint_receipt(
    pos_id,                     # type: str
    receipts,                   # type: List[str]
    fiscalization_type="",      # type: Optional[str]
    error_title="$EFT",         # type: Optional[str]
):
    # type: (...) -> bool

    reprinted = True
    for receipt in receipts:
        eft_check = _check_eft_value(
            pos_id=pos_id,
            eft_value=receipt,
            error_message="$PAYMENT_HAS_NO_TEF",
            error_title=error_title,
        )
        if not eft_check:
            reprinted = False
            continue

        printer_receipt = PrintReceipt(
            mb_context=mbcontext,
            logger=logger,
        )

        reprinted &= printer_receipt.reprint_receipt(
            pos_id=pos_id,
            receipts=[receipt],
            fiscalization_type=fiscalization_type,
        )

    return reprinted


def _check_eft_value(
    pos_id,                 # type: str
    eft_value,              # type: str
    error_message,          # type: str
    error_title="#$EFT",    # type: Optional[str]
):
    # type (...) -> bool

    if eft_value in ['', None]:
        sysactions.show_messagebox(
            posid=pos_id,
            message=error_message,
            title=error_title,
            buttons="$OK",
        )
        return False

    return True



from ._CloseDialogProcessor import ElectronicPaymentCloseDialogProcessor
from ._ShowQRCodeProcessor import ElectronicPaymentShowQRCodeProcessor
from ._ShowChoicesProcessor import ElectronicPaymentShowChoicesProcessor
from ._ShowInputProcessor import ElectronicPaymentShowInputProcessor
from ._ShowMessageProcessor import ElectronicPaymentShowMessageProcessor
from ._ShowQuestionProcessor import ElectronicPaymentShowQuestionProcessor



from helper import ExtendedEnum
from ._LoyaltyToken import LoyaltyToken


class ResponseToken(ExtendedEnum):
    TK_LOYALTY_ID_NOT_FOUND = LoyaltyToken.create_token(number="F001")
    TK_LOYALTY_ERROR = LoyaltyToken.create_token(number="F002")



# encoding: utf-8

from typing import List

from old_helper import convert_from_utf_to_localtime
from helper import remove_accents, format_cpf_cnpj
from systools import sys_log_exception
from typing import (
    List,
    Any,
    Dict,
)

from basereport import (
    TableReport,
    ReportColumnDefinition,
    SimpleReport,
    AlignTypes,
)
from report import Presenter, Part
from report.command import NewLineCommand, AlignCommand, RepeatCommand
from utils.report_formatters import set_font, rollback_printer_default_font
from commons.util import LogoImage
from dto import DeliveryBodyDto


class DeliveryPresenter(Presenter):
    def __init__(self, columns):
        self.columns = columns

        self.header_column_definition = [
            ReportColumnDefinition(width=8, after_fill_text=': ', fill_with_char='.'),
            ReportColumnDefinition(width=columns - 8)
        ]

        self.info_value_column_definition = [
            ReportColumnDefinition(width=columns - 8),
            ReportColumnDefinition(width=8)
        ]

        self.delivery_info_column_definition = [
            ReportColumnDefinition(width=13, after_fill_text=': ', fill_with_char='.'),
            ReportColumnDefinition(width=columns - 13)
        ]

        self.header_table_column_definitions = [
            ReportColumnDefinition(
                width=4,
            ),
            ReportColumnDefinition(
                width=columns - 13,
            ),
            ReportColumnDefinition(
                width=9,
                before_text=" ",
                align=AlignTypes.RIGHT,
            ),
        ]

        self.observation_column_definition = [
            ReportColumnDefinition(width=13, after_fill_text=': ', fill_with_char='.'),
            ReportColumnDefinition(width=columns - 13)
        ]

        self.full_line_column_definitions = [
            ReportColumnDefinition(width=columns),
            ReportColumnDefinition(width=0)
        ]

    def present(self, dto):
        try:
            parts = set_font("B")
            parts.extend(self._build_header(dto.header))
            parts.extend(self._build_delivery_information(
                body=dto.body,
                show_customer_phone=dto.show_customer_phone,
            ))
            parts.extend(self._build_items(dto.body, dto.products, dto.cfh_products))
            parts.extend(self._build_order_values(dto.body))
            parts.extend(self._build_payments(dto.body))
            parts.extend(self._build_observation(dto.body))
            parts.extend(self._build_logo())
            parts.extend([Part(rollback_printer_default_font(), [NewLineCommand()])])
        except Exception as _:
            sys_log_exception("Error building Delivery Report")
            raise

        return SimpleReport(parts, self.columns)

    def _build_header(self, header):
        parts = [
            Part(None, [NewLineCommand()]),
            self._get_store_name(header),
            Part(None, [NewLineCommand()]),
            Part(None, [NewLineCommand()])
        ]
        parts.extend(self._get_order_number(header))
        parts.extend(self._get_order_date(header))

        return parts

    def _build_items(self, body, products, cfh_products):
        items = [
            Part("$ITEMS_DELIVERY_REPORT", [AlignCommand(AlignCommand.Alignment.center)]),
            Part(None, [NewLineCommand()]),
            Part(None, [RepeatCommand(self.columns, "#"), NewLineCommand()]),
            TableReport([["$QTY_DELIVERY_REPORT", "$DESCRIPTION_DELIVERY_REPORT", "$VALUE_DELIVERY_REPORT"]],
                        self.header_table_column_definitions),
            Part(None, [RepeatCommand(self.columns, "#"), NewLineCommand()])
        ]

        for item in body.items:
            items.extend(
                    self._build_item(
                        item=item,
                        products=products,
                        cfh_products=cfh_products,
                        level=0,
                    )
            )

        items.append(Part(None, [RepeatCommand(self.columns, "-"), NewLineCommand()]))

        return items

    def _build_item(
        self,
        item,               # type: Dict[Any]
        products,           # type: Dict[Any]
        cfh_products,       # type: Dict[Any]
        level=0,            # type: int
    ):
        # type: (...) -> List[TableReport]

        items = []

        qty = int(item["quantity"])
        item_quantity = unicode(qty)
        item_part_code = unicode(item["partCode"])
        product_name = self._get_product_name(item, products, item_part_code)
        product_type = products[item_part_code][1] if item_part_code in products else ''
        item_name = remove_accents(product_name)
        item_price = float(item["price"])
        item_formatted_price = self._format_price(item_price) if item_price > 0 else ''
        item_line = [item_quantity, item_name, item_formatted_price]
        item_type = item["itemType"] if "itemType" in item else product_type

        increment_level = 0
        if (item_price > 0 or item_type != "OPTION") and item_part_code not in cfh_products:
            increment_level = 1
            column_definition = self._get_column_definition(
                level=level,
                item_formatted_price=item_formatted_price,
            )

            items.append(TableReport([item_line], column_definition))

            if "observation" in item and item["observation"] is not None and item["observation"] != "":
                item_observation = "   " + item[u"observation"]
                items.extend(self._format_line('', item_observation, ''))

        if "parts" in item:
            for sub_item in item["parts"]:
                items.extend(
                    self._build_item(
                        item=sub_item,
                        products=products,
                        cfh_products=cfh_products,
                        level=level+increment_level,
                    )
                )

        return items

    def _get_column_definition(
        self,
        level,                  # type: int
        item_formatted_price,   # type: str
    ):
        # type: (...) -> List[ReportColumnDefinition]

        before_text = (" " * (level * 2)) if level > 0 else ""
        before_text_column_width = (len(before_text) + 4)
        after_text_column_width = 9 if item_formatted_price != '' else 0

        return [
            ReportColumnDefinition(
                width=before_text_column_width,
                before_text=before_text,
            ),
            ReportColumnDefinition(
                width=self.columns - (before_text_column_width + after_text_column_width),
                before_text=" ",
            ),
            ReportColumnDefinition(
                width=after_text_column_width,
                before_text=" ",
            ),
        ]

    @staticmethod
    def _get_product_name(item, products, item_part_code):
        product_name = item.get("name", None)
        if not product_name:
            product_name = products[item_part_code][0] if item_part_code in products else item_part_code
        return product_name

    def _build_order_values(self, body):
        values = []

        item_line = ["$TOTAL_ITEMS_DELIVERY_REPORT",  self._format_price(float(body.sub_total_price))]
        values.append(TableReport([item_line], self.info_value_column_definition))

        delivery_fee = float(body.delivery_fee)
        if delivery_fee > 0:
            item_line = ["$DELIVERY_FEE_DELIVERY_REPORT", self._format_price(float(delivery_fee))]
            values.append(TableReport([item_line], self.info_value_column_definition))

        additional_fee = float(body.additional_fee)
        if additional_fee > 0:
            item_line = ["$DELIVERY_ADDITIONAL_FEE_DELIVERY_REPORT", self._format_price(float(additional_fee))]
            values.append(
                TableReport(
                    lines=[item_line],
                    column_definitions=self.info_value_column_definition,
                )
            )

        if body.vouchers:
            for key, value in body.vouchers.items():
                if float(value) > 0:
                    item_line = ["$" + key + "_DISCOUNT_DELIVERY_REPORT", self._format_price(float(value))]
                    values.append(TableReport([item_line], self.info_value_column_definition))

        item_line = ["$TOTAL_DELIVERY_REPORT", self._format_price(float(body.total_price))]
        values.append(TableReport([item_line], self.info_value_column_definition))

        values.append(Part(None, [RepeatCommand(self.columns, "-"), NewLineCommand()]))

        return values

    def _build_payments(self, body):
        tenders = [
            Part("$PAYMENT_REPORT", [AlignCommand(AlignCommand.Alignment.center)]),
            Part(None, [NewLineCommand()]),
            Part(None, [RepeatCommand(self.columns, "-"), NewLineCommand()])
        ]

        for tender in body.tenders:
            item_line = [tender["type"], self._format_price(float(tender["value"]))]
            tenders.append(TableReport([item_line], self.info_value_column_definition))

        tenders.append(Part(None, [RepeatCommand(self.columns, "-"), NewLineCommand()]))

        return tenders

    def _build_delivery_information(
        self,
        body,                   # type: DeliveryBodyDto
        show_customer_phone,    # type: bool
    ):
        # type: (...) -> List[Part]

        delivery_information = [
            Part(None, [NewLineCommand()]),
            Part("$INFORMATION_DELIVERY_REPORT", [AlignCommand(AlignCommand.Alignment.center)]),
            Part(None, [NewLineCommand()]),
            Part(None, [RepeatCommand(self.columns, "-"), NewLineCommand()])
        ]

        delivery_information.extend(self._format_line("$CUSTOMER_NAME_DELIVERY_REPORT",
                                                      body.customer_name,
                                                      self.delivery_info_column_definition))

        if show_customer_phone and (body.partner.lower() != "ifood" or body.need_logistics):
            delivery_information.extend(self._format_line(
                title="$PHONE_DELIVERY_REPORT",
                text=body.customer_phone,
                definition=self.delivery_info_column_definition,
            ))

        delivery_information.extend(self._format_line("$ADDRESS_DELIVERY_REPORT",
                                                      body.address,
                                                      self.delivery_info_column_definition))

        delivery_information.extend(self._format_line("$COMPLEMENT_DELIVERY_REPORT",
                                                      body.address_complement,
                                                      self.delivery_info_column_definition))

        delivery_information.extend(self._format_line("$NEIGHBORHOOD_DELIVERY_REPORT",
                                                      body.address_neighborhood,
                                                      self.delivery_info_column_definition))
        delivery_information.extend(self._format_line("$CITY_DELIVERY_REPORT",
                                                      body.city,
                                                      self.delivery_info_column_definition))
        delivery_information.extend(self._format_line("$POSTAL_CODE_DELIVERY_REPORT",
                                                      body.postal_code,
                                                      self.delivery_info_column_definition))

        delivery_information.extend([
            Part(None, [RepeatCommand(self.columns, "-"), NewLineCommand()]),
            Part(None, [NewLineCommand()])
        ])

        return delivery_information

    def _build_observation(self, body):
        if not body.observation and body.observations == "" and not body.customer_doc:
            return []

        observations = [
            Part("$OBSERVATION_DELIVERY_REPORT", [AlignCommand(AlignCommand.Alignment.center)]),
            Part(None, [NewLineCommand()]),
            Part(None, [RepeatCommand(self.columns, "-")]),
            Part(None, [NewLineCommand()])
        ]

        if body.observation:
            for observation in body.observation.split(";"):
                observations.extend(self._format_line("", observation.strip(), self.full_line_column_definitions))

        if body.customer_doc:
            title = "$DOCUMENT_DELIVERY_REPORT"
            text = format_cpf_cnpj(body.customer_doc)
            definition = self.observation_column_definition
            observations.extend([Part(None, [NewLineCommand()])])
            observations.extend(self._format_line(title, text, definition))

        if body.observations != "":
            title = ""
            text = body.observations
            definition = self.observation_column_definition
            observations.extend([Part(None, [NewLineCommand()])])
            observations.extend(self._format_line(title, text, definition))

        observations.append(Part(None, [RepeatCommand(self.columns, "-")]))
        return observations

    @staticmethod
    def _build_developer_info():
        developer_info = [
            Part(None, [NewLineCommand()]),
            Part(None, [NewLineCommand()]),
            Part("$DEVELOPER_DELIVERY_REPORT", [AlignCommand(AlignCommand.Alignment.center)]),
            Part(None, [NewLineCommand()]),
            Part("$DEVELOPER_WEB_SITE_DELIVERY_REPORT", [AlignCommand(AlignCommand.Alignment.center)]),
            Part(None, [NewLineCommand()]),
            Part(None, [NewLineCommand()]),
            Part(None, [NewLineCommand()]),
            Part(None, [NewLineCommand()])
        ]

        return developer_info

    @staticmethod
    def _build_logo():
        logo = LogoImage().get_logo()
        developer_info = [
            Part(None, [NewLineCommand()]),
            Part(None, [NewLineCommand()]),
            Part(logo.strip(), [AlignCommand(AlignCommand.Alignment.center)]),
            Part("$DEVELOPER_WEB_SITE_DELIVERY_REPORT_3S", [AlignCommand(AlignCommand.Alignment.center)]),
            Part(None, [NewLineCommand()]),
            Part(None, [NewLineCommand()]),
            Part(None, [NewLineCommand()]),
        ]

        return developer_info

    def _format_line(self, title, text, definition):
        if not text:
            return []

        lines = []
        current_string_start = 0
        current_string_end = definition[1].width if title else self.columns

        while len(text) > 0:
            if title and current_string_start == 0:
                lines.append(TableReport([[title, text[:current_string_end]]], definition))
            else:
                lines.append(TableReport([[text[current_string_start:current_string_end]]],
                                         [ReportColumnDefinition(width=self.columns)]))

            if current_string_end > len(text):
                break

            current_string_start = current_string_end
            current_string_end += self.columns

        return lines

    @staticmethod
    def _get_store_name(header):
        return Part(header.store_name.upper(), [AlignCommand(AlignCommand.Alignment.center)])

    def _get_order_number(self, header):
        return self._format_line("$ORDER_DELIVERY_REPORT", header.order_number, self.header_column_definition)

    def _get_order_date(self, header):
        if header.partner.lower() != "delivery":
            created_at = convert_from_utf_to_localtime(header.order_date).strftime("%d/%m/%Y %H:%M:%S")
        else:
            created_at = header.order_date.strftime("%d/%m/%Y %H:%M:%S")

        return self._format_line("$DATE_DELIVERY_REPORT", created_at, self.header_column_definition)

    @staticmethod
    def _format_price(price):
        formatted_value = '{:8,.2f}'.format(price)
        return formatted_value



import datetime
import time

import sysactions
from actions.void.utils import void_order
from posot import OrderTakerException


def do_recall_order(pos_id, order_id, originator_pos_id=None, check_date=True, pos_ot=None):
    model = sysactions.get_model(pos_id)
    if pos_ot is None:
        pos_ot = sysactions.get_posot(model)
    session = sysactions.get_operator_session(model)

    if originator_pos_id:
        recall_pos = int(originator_pos_id[4:])
        try:
            pos_ot.recallOrder(int(pos_id), int(order_id), session, sourceposid=str(recall_pos))
        except OrderTakerException as _:
            pos_ot.recallOrder(int(pos_id), int(order_id), session)
    else:
        pos_ot.recallOrder(int(pos_id), int(order_id), session)

    if check_date is True:
        # Waits a maximum of 5 seconds for the order to "arrive" at the POS model
        time_limit = (time.time() + 5.0)
        while (time.time() < time_limit) and (not sysactions.has_current_order(sysactions.get_model(pos_id))):
            time.sleep(0.1)

        model = sysactions.get_model(pos_id)
        order = model.find("CurrentOrder/Order")

        # Cannot recall order older than 20 hours
        now = datetime.datetime.now()
        order_created_date = datetime.datetime.strptime(order.get("createdAt"), "%Y-%m-%dT%H:%M:%S.%f")
        if order_created_date < now - datetime.timedelta(hours=20):
            sysactions.show_messagebox(pos_id, "$CANNOT_RECALL_ORDER_TOO_OLD", title="$ERROR", icon="error")
            void_order(
                pos_id=pos_id,
                void_reason="RECALL_ORDER_TO_OLD",
                pos_ot=pos_ot,
                void_type="VOID_STORED_ORDER",
            )



from typing import List, Dict


class ActionResult(object):
    def __init__(self, discounts_id, items_to_burn, order_tender_ids, discount_item_values, benefit_applied,
                 pending_to_apply):
        # type: (List, Dict, List, Dict, bool, bool) -> None

        self.discounts_id = discounts_id
        self.items_to_burn = items_to_burn
        self.order_tenders = order_tender_ids
        self.discount_item_values = discount_item_values
        self.benefit_applied = benefit_applied
        self.pending_to_apply = pending_to_apply

    def __add__(self, action_result):
        # type: (ActionResult) -> ActionResult

        return self._merge_action_result(self, action_result)

    def __radd__(self, action_result):
        # type: (ActionResult) -> ActionResult

        return self._merge_action_result(self, action_result)

    def __nonzero__(self):
        # type: () -> bool
        return bool(self.discounts_id) or bool(self.items_to_burn) or bool(self.order_tenders) or self.pending_to_apply

    @staticmethod
    def _merge_action_result(current, new):
        current.discounts_id.extend(new.discounts_id)
        current.order_tenders.extend(new.order_tenders)
        current.discount_item_values.update(new.discount_item_values)
        current.benefit_applied = current.benefit_applied and new.benefit_applied
        current.pending_to_apply = current.pending_to_apply or new.pending_to_apply

        return ActionResult(current.discounts_id, current.items_to_burn, current.order_tenders,
                            current.discount_item_values, current.benefit_applied, current.pending_to_apply)




from typing import List

from basereport import TableReport, ReportColumnDefinition, AlignTypes
from report import Presenter, Part
from report.command import RepeatCommand
from ..dto import TipReportBodyDto


class TipReportBodyPresenter(Presenter):
    def present(self, dto):
        # type: (TipReportBodyDto) -> List[TableReport]
        report_body = []

        detail_lines = [["$TIP_REPORT_OPERATOR", "$TIP_REPORT_TABLES", "$TIP_REPORT_ORDERS", "#CURRENCY_SYMBOL()", "$TIP_REPORT_TIP"]]
        for detail_line in dto.detail_line_list:
            detail_lines.append([
                detail_line.operator,
                detail_line.table_count,
                detail_line.order_count,
                "#NUMBER({})".format(detail_line.total_sold_amount),
                "#NUMBER({})".format(detail_line.tip_amount)
            ])

        report_body.append(TableReport(lines=detail_lines, column_definitions=a_item_column_definition()))

        report_body.extend([Part(commands=[RepeatCommand(38, "=")]), a_new_line()])

        totalized_line = [
            [
            "$TOTAL",
            dto.total_tables_count,
            dto.total_order_count,
            "#NUMBER({})".format(dto.total_sold_amount),
            "#NUMBER({})".format(dto.total_tip_amount)
            ]
        ]

        report_body.append(TableReport(lines=totalized_line, column_definitions=a_item_column_definition()))

        return report_body

def a_new_line():
    return Part("\n")


def a_item_column_definition():
    return [
        ReportColumnDefinition(width=8,
                               before_text="",
                               after_text="",
                               after_fill_text="",
                               align=AlignTypes.CENTER),
        ReportColumnDefinition(width=7,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.CENTER),
        ReportColumnDefinition(width=7,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.CENTER),
        ReportColumnDefinition(width=8,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.CENTER),
        ReportColumnDefinition(width=8,
                               fill_with_char="",
                               before_text="",
                               align=AlignTypes.CENTER)
    ]



from commons.dto import DefaultHeaderDto
from commons.report import Generator
from reports_app.cashtransferreport.dto import CashTransferReportDto  # noqa
from reports_app.cashtransferreport.dto import CashReportFiltersDto  # noqa
from reports_app.cashtransferreport.dto import TransfersByDayDto


class CashTransferGenerator(Generator):
    def __init__(self, account_repository, cash_transfer_filter):
        # type: (AccountRepository, CashReportFiltersDto) -> None
        self.account_repository = account_repository
        self.cash_transfer_filter = cash_transfer_filter

    def generate_data(self):
        header_transfer_dto = DefaultHeaderDto(
            self.cash_transfer_filter.store_id,
            self.cash_transfer_filter.pos_id,
            self.cash_transfer_filter.operator_id,
            self.cash_transfer_filter.initial_date.strftime("%d/%m/%Y") + " - " +
            self.cash_transfer_filter.end_date.strftime("%d/%m/%Y"),
            self.cash_transfer_filter.report_type)

        transfers = self.account_repository.get_transfers_by_business_period(
            self.cash_transfer_filter.initial_date,
            self.cash_transfer_filter.end_date,
            self.cash_transfer_filter.operator_id,
            self.cash_transfer_filter.pos_id,
            self.cash_transfer_filter.report_type)

        all_transfers_dto = self._append_transfer_in_transfers_list(transfers)

        return CashTransferReportDto(header_transfer_dto, all_transfers_dto, self.cash_transfer_filter.caller_pos_id)

    def _append_transfer_in_transfers_list(self, transfers):
        all_transfers_dto = []
        for transfer in transfers:
            transfer_date = self._get_datetime_date(transfer.transfer_date)
            element_found = self._if_transfer_day_exists_insert_transfer(all_transfers_dto, transfer, transfer_date)
            self._create_new_transfer_day(all_transfers_dto, element_found, transfer, transfer_date)

        return all_transfers_dto

    @staticmethod
    def _create_new_transfer_day(all_transfers_dto, element_found, transfer, transfer_date):
        if not element_found:
            all_transfers_dto.append(TransfersByDayDto(
                transfer_date, [transfer]
            ))

    @staticmethod
    def _if_transfer_day_exists_insert_transfer(all_transfers_dto, transfer, transfer_date):
        element_found = False

        for transfer_by_day in all_transfers_dto:
            if transfer_by_day.day == transfer_date:
                transfer_by_day.transfers_by_date.append(transfer)
                element_found = True
                break
        return element_found

    @staticmethod
    def _get_datetime_date(time):
        return time.strftime("%d/%m/%Y")



# encoding: utf-8

from old_helper import BaseRepository


class OrderRepository(BaseRepository):

    def __init__(self, mbcontext, pos_list):
        super(OrderRepository, self).__init__(mbcontext)
        self.pos_list = pos_list


    def get_paid_orders(self, business_period):
        def inner_func(conn):
            query = self._PaidOrdersQuery.format(business_period)
            orders = [[(x.get_entry(0), x.get_entry(1)) for x in conn.select(query)][0]]
            return orders

        return self.execute_parallel_in_all_databases_returning_flat_list(inner_func, self.pos_list)

    _PaidOrdersQuery = """\
SELECT SessionId, COUNT(*) AS orders
FROM orders
WHERE StateId = 5 AND BusinessPeriod = '{}'"""


# ASN.1 types tags
from operator import getitem
from pyasn1 import error

tagClassUniversal = 0x00
tagClassApplication = 0x40
tagClassContext = 0x80
tagClassPrivate = 0xC0

tagFormatSimple = 0x00
tagFormatConstructed = 0x20

tagCategoryImplicit = 0x01
tagCategoryExplicit = 0x02
tagCategoryUntagged = 0x04

class Tag:
    def __init__(self, tagClass, tagFormat, tagId):
        if tagId < 0:
            raise error.PyAsn1Error(
                'Negative tag ID (%s) not allowed' % (tagId,)
                )
        self.__tag = (tagClass, tagFormat, tagId)
        self.uniq = (tagClass, tagId)
        self.__hashedUniqTag = hash(self.uniq)

    def __str__(self):
        return '[%s:%s:%s]' % self.__tag

    def __repr__(self):
        return '%s(tagClass=%s, tagFormat=%s, tagId=%s)' % (
            (self.__class__.__name__,) + self.__tag
            )
    # These is really a hotspot -- expose public "uniq" attribute to save on
    # function calls
    def __eq__(self, other): return self.uniq == other.uniq
    def __ne__(self, other): return self.uniq != other.uniq
    def __lt__(self, other): return self.uniq < other.uniq
    def __le__(self, other): return self.uniq <= other.uniq
    def __gt__(self, other): return self.uniq > other.uniq
    def __ge__(self, other): return self.uniq >= other.uniq
    def __hash__(self): return self.__hashedUniqTag
    def __getitem__(self, idx): return self.__tag[idx]
    def __and__(self, otherTag):
        (tagClass, tagFormat, tagId) = otherTag
        return self.__class__(
            self.__tag&tagClass, self.__tag&tagFormat, self.__tag&tagId
            )
    def __or__(self, otherTag):
        (tagClass, tagFormat, tagId) = otherTag
        return self.__class__(
            self.__tag[0]|tagClass,
            self.__tag[1]|tagFormat,
            self.__tag[2]|tagId
            )
    def asTuple(self): return self.__tag  # __getitem__() is slow
    
class TagSet:
    def __init__(self, baseTag=(), *superTags):
        self.__baseTag = baseTag
        self.__superTags = superTags
        self.__hashedSuperTags = hash(superTags)
        _uniq = ()
        for t in superTags:
            _uniq = _uniq + t.uniq
        self.uniq = _uniq
        self.__lenOfSuperTags = len(superTags)

    def __str__(self):
        return self.__superTags and '+'.join([str(x) for x in self.__superTags]) or '[untagged]'
 
    def __repr__(self):
        return '%s(%s)' % (
            self.__class__.__name__,
            '(), ' + ', '.join([repr(x) for x in self.__superTags])
            )

    def __add__(self, superTag):
        return self.__class__(
            self.__baseTag, *self.__superTags + (superTag,)
            )
    def __radd__(self, superTag):
        return self.__class__(
            self.__baseTag, *(superTag,) + self.__superTags
            )

    def tagExplicitly(self, superTag):
        tagClass, tagFormat, tagId = superTag
        if tagClass == tagClassUniversal:
            raise error.PyAsn1Error(
                'Can\'t tag with UNIVERSAL-class tag'
                )
        if tagFormat != tagFormatConstructed:
            superTag = Tag(tagClass, tagFormatConstructed, tagId)
        return self + superTag

    def tagImplicitly(self, superTag):
        tagClass, tagFormat, tagId = superTag
        if self.__superTags:
            superTag = Tag(tagClass, self.__superTags[-1][1], tagId)
        return self[:-1] + superTag

    def getBaseTag(self): return self.__baseTag
    def __getitem__(self, idx):
        if isinstance(idx, slice):
            return self.__class__(
               self.__baseTag, *getitem(self.__superTags, idx)
            )
        return self.__superTags[idx]
    def __eq__(self, other): return self.uniq == other.uniq
    def __ne__(self, other): return self.uniq != other.uniq
    def __lt__(self, other): return self.uniq < other.uniq
    def __le__(self, other): return self.uniq <= other.uniq
    def __gt__(self, other): return self.uniq > other.uniq
    def __ge__(self, other): return self.uniq >= other.uniq
    def __hash__(self): return self.__hashedSuperTags
    def __len__(self): return self.__lenOfSuperTags
    def isSuperTagSetOf(self, tagSet):
        if len(tagSet) < self.__lenOfSuperTags:
            return
        idx = self.__lenOfSuperTags - 1
        while idx >= 0:
            if self.__superTags[idx] != tagSet[idx]:
                return
            idx = idx - 1
        return 1
    
def initTagSet(tag): return TagSet(tag, tag)
